package tetasarts.generator;


import java.util.*;
import java.util.logging.Logger;

import org.apache.bcel.Repository;
import org.apache.bcel.classfile.*;

import tetasarts.Options;
import tetasarts.tir.IntermediateData;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.methods.LazyMethodInformation;
import tetasarts.tir.methods.MethodInformation;

public class JavaTranslator {
	protected Logger logger = Logger.getLogger("generator"); 
	protected String mainClass;
	protected IntermediateData intermediateData;
	protected HashSet<String> knownClasses;
	protected AbstractClassNode currentClassTarget;
	protected LocationBuilder locationBuilder;
	protected TransitionBuilder transitionBuilder;
	protected Map<Integer, AbstractBasicBlock> instructionNodeMap;
	protected ClassGraph classGraph;
	protected ClassGraphBuilder classGraphBuilder;

	public HashSet<String> getKnownClasses() {
		return knownClasses;
	}
	
	public IntermediateData getIntermediateData() {
		return intermediateData;
	}
	
	public ClassGraph getClassGraph() {
		return this.classGraph;
	}
	
	public String getMainClass() {
		return this.mainClass;
	}
	
	public JavaTranslator(String classname) {
		knownClasses = new HashSet<String>();
		intermediateData = new IntermediateData();
		classGraph = intermediateData.getClassGraph();

		mainClass = classname;

		locationBuilder = new LocationBuilder(this);
		transitionBuilder = new TransitionBuilder(this);
		classGraphBuilder = new ClassGraphBuilder();
	}

	public IntermediateData buildIR() {
		
		System.out.println("build jop classes");
		if(Options.USE_JOP_OPCODE_METHODS)
			classGraphBuilder.buildClassGraph(Constants.JOP_JVM_CLASS, classGraph);
//		classGraphBuilder.buildClassGraph("javax.scj.PeriodicThread", classGraph);
		//System.out.println("PeriodicThread done");
//		System.out.println("PeriodicThread done");
		//System.out.println("build mainclass");
		classGraphBuilder.buildClassGraph(mainClass, classGraph);
		System.out.println("Building class graph done");
		
		intermediateData.setMainClass(classGraph.getClassFromCanonicalName(mainClass));

		for (AbstractClassNode classNode : classGraph.getNodes()) {
			buildClass(classNode);
		}

		return intermediateData;
	}
	
	protected void buildClass(AbstractClassNode classNode) {
		
		currentClassTarget = classNode;
		JavaClass jc = classNode.getJavaClass();
		
		if(jc == null) {
			return;
		}
		
		locationBuilder.setCurrentClass(jc);
		transitionBuilder.setCurrentClass(jc);
		
		buildMethods(jc);
	}
	
	protected void buildMethods(JavaClass jc) {
		Method[] methods = jc.getMethods();
		HashMap<Method, MethodInformation> methodToMethodInfoMap = new HashMap<Method, MethodInformation>();
		LoopBoundAnalysis loopboundanalysis;
		for (Method currentMethod : methods) {
			if (currentMethod.isAbstract()) { // skip abstract methods
				continue;
			}
			if (currentMethod.isNative()) { // skip native methods
				continue;
			}

			MethodInformation currentMethodTarget;
			if(Options.LAZY_METHODS){ // are we using lazy methodloading?
				currentMethodTarget = new LazyMethodInformation(jc, currentMethod, locationBuilder, transitionBuilder);
			} else {
				currentMethodTarget = new MethodInformation(jc.getClassName(), currentMethod.getName(), currentMethod.getSignature(), currentMethod.getCode().getCode().length);
			}
			currentMethodTarget.setUsed(forceGenerate(currentMethodTarget)); // we force generation of some methods
			currentMethodTarget.setSynchronized(currentMethod.isSynchronized());

			currentClassTarget.addMethod(currentMethodTarget);
			
			instructionNodeMap = new HashMap<Integer, AbstractBasicBlock>();
			if(! Options.LAZY_METHODS) {
				locationBuilder.visitMethod(currentMethod, currentMethodTarget);

				transitionBuilder.visitMethod(currentMethod, currentMethodTarget);
				if(Options.PRECISE_LOOP) {
					loopboundanalysis = new LoopBoundAnalysis(currentMethodTarget.getInstructionGraph(), currentMethod, jc);
					loopboundanalysis.conductAnalysis();
				}
			}
		}
		
	}
	
	
	/**
	 * Method for determining if this method should be generated even if it's not used explicitly. This is relevant for e.g. finalizers.
	 * @param mi - MethodInformation - SIR representation of method.
	 * @return true/false yes/no
	 */
	private boolean forceGenerate(MethodInformation mi){
		boolean result= false;
		
		result |= ((Options.CONSIDER_FINALIZERS)&&((mi.getName()+mi.getSignature()).equals("finalize()V"))); // generate finalizers
		
//		if(mi.getName().equals("exit"))
//			result = true;

		
		return result;
	}
}
