package tetasarts.generator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.util.ClassPath.ClassFile;

import tetasarts.ClassLookup;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.LoopBound;
import tetasarts.tir.basicblocks.basicBlockTypes.LoopBasicBlock;

public class LoopBoundAnalysis {

	private BasicBlockGraph bbGraph;
	private int globalBasicBlockCount;
	private HashSet<AbstractBasicBlock> visitedBBs;
	private HashMap<AbstractBasicBlock, Integer> bbToDfstValMap;
	private LinkedList<BackEdge> backEdges;
	private HashMap<AbstractBasicBlock, LoopInfo> loopHeadersToInfoMap;
	private Method currentMethod;
	private JavaClass currentClass;
	
	private HashMap<Integer, String> sourceLineMap;
	private Pattern loopBoundPattern = Pattern.compile("([0-9]+)");
	
	private class BackEdge {
		private AbstractBasicBlock src;
		private AbstractBasicBlock dest;
		public BackEdge(AbstractBasicBlock source, AbstractBasicBlock destination) {
			this.src = source;
			this.dest = destination;
		}
		public AbstractBasicBlock getSrc() {
			return this.src;
		}
		public AbstractBasicBlock getDest() {
			return this.dest;
		}
		
		public int hashCode() {
	        return src.hashCode() + dest.hashCode();
	    }
		@Override
	    public boolean equals(Object obj) {
	        if (obj == null)
	            return false;
	        if (obj == this)
	            return true;
	        if (obj.getClass() != getClass())
	            return false;
			
	        BackEdge rhs = (BackEdge) obj;
	        
	        if(rhs.getSrc().getID() == src.getID() &&
	           rhs.getDest().getID() == dest.getID())
	        	return true;
	        else
	        	return false;
	    }
		
	}
	
	private class LoopInfo {
		private int loopBound;
		private LinkedList<AbstractBasicBlock> bbs;
		private LinkedList<BackEdge> backEdges;
		
		public LoopInfo() {
			this.loopBound = -1;
			this.bbs = new LinkedList<AbstractBasicBlock>();
			this.backEdges = new LinkedList<LoopBoundAnalysis.BackEdge>();
		}
		
		public void setLoopBound(int bound) {
			this.loopBound = bound;
		}
		
		public int getLoopBound() {
			return this.loopBound;
		}
		
		public void addBasicBlock(AbstractBasicBlock basicBlock) {
			if(!this.bbs.contains(basicBlock))
				this.bbs.add(basicBlock);
		}
		
		public LinkedList<AbstractBasicBlock> getBasicBlocks() {
			return this.bbs;
		}
		
		public void addBackEdge(BackEdge edge) {
			this.backEdges.add(edge);
		}
		
		public LinkedList<BackEdge> getBackEdges() {
			return this.backEdges;
		}
		
		public AbstractBasicBlock getLoopHeader(BasicBlockGraph graph) {
			return findConditionalBlock(this.backEdges.getFirst().getDest(), graph, this.bbs);
		}
		
		private AbstractBasicBlock findConditionalBlock(AbstractBasicBlock bb, BasicBlockGraph graph, LinkedList<AbstractBasicBlock> loopBBs) {
			if(bb.getBytecodes().get(0).getInstruction() instanceof BranchInstruction)
				return bb;
			List<AbstractBasicBlock> succs = graph.getSuccs(bb);
			for(AbstractBasicBlock succ : succs) {
				if(loopBBs.contains(succ)) {
					AbstractBasicBlock cond = findConditionalBlock(succ, graph, loopBBs);
					if(cond != null)
						return cond;
				}
			}
			return null;
		}
	}
	
	
	public LoopBoundAnalysis(BasicBlockGraph graph, Method analysedMethod, JavaClass analysedClass) {
		this.bbGraph = graph;
		this.globalBasicBlockCount = bbGraph.getNodes().size();
		this.visitedBBs = new HashSet<AbstractBasicBlock>();
		this.bbToDfstValMap = new HashMap<AbstractBasicBlock, Integer>();
		this.backEdges = new LinkedList<LoopBoundAnalysis.BackEdge>();
		this.loopHeadersToInfoMap = new HashMap<AbstractBasicBlock, LoopInfo>();
		
		//we can probably delete these two
		this.currentMethod = analysedMethod;
		this.currentClass = analysedClass;
		
		this.sourceLineMap = getSourceLineMap(analysedClass);	
	}
	
	public void conductAnalysis() {
		List<AbstractBasicBlock> heads = this.bbGraph.getHeads();
		assert heads.size() > 1 : "multiple heads!";
		
		depthFirstOrdering(heads.get(0));
		this.visitedBBs.clear();
		searchForBackEdges(heads.get(0));

		LinkedList<LoopInfo> foundLoops = new LinkedList<LoopInfo>();		
		for(BackEdge e : this.backEdges) {
			this.visitedBBs.clear();
			constructLoopSet(e, foundLoops);
		}
		
		for(LoopInfo info : foundLoops) {
			setLoopBound(info);
		}
		
		//hooking up to original solution.... This is pretty ugly..
		for(LoopInfo info : foundLoops) {
			AbstractBasicBlock branchingBB = info.getLoopHeader(bbGraph);
			
			AbstractBasicBlock targetbb = null;
			for(AbstractBasicBlock branchSucc : bbGraph.getSuccs(branchingBB)) {
				if(info.getBasicBlocks().contains(branchSucc)){
					targetbb = branchSucc;
					break;
				}
			}
			
			
			/*if(branchingBB.getBytecodes().get(0).getInstruction()
			   instanceof BranchInstruction) {
				BranchInstruction branch = (BranchInstruction)branchingBB.getBytecodes().get(0).getInstruction();
				InstructionHandle targetIns = branch.getTarget();
				for(AbstractBasicBlock b : bbGraph.getNodes()) {
					for(InstructionHandle handle : b.getBytecodes()) {
						if(handle.getPosition() == targetIns.getPosition()) {
							targetbb = b;
						}
					}
				}
			}
			*/
			
			targetbb.setTarget(true);
			LoopBound lb = new LoopBound(targetbb, info.getLoopBound());
			AbstractBasicBlock newbb = new LoopBasicBlock(lb);
			newbb.setID(branchingBB.getID());
			lb.addLoopNode((LoopBasicBlock)newbb);
			
			List<AbstractBasicBlock> succs = bbGraph.getSuccs(branchingBB);
			List<AbstractBasicBlock> preds = bbGraph.getPreds(branchingBB);
			List<InstructionHandle> bytecodes = branchingBB.getBytecodes();
			
			
			
			
			bbGraph.addNode(newbb);
			for(AbstractBasicBlock predbb : preds) {
				//bbGraph.addEdge(newbb, predbb);
				bbGraph.addEdge(predbb, newbb);
			}
			for(AbstractBasicBlock succbb : succs) {
				bbGraph.addEdge(newbb, succbb);
				//bbGraph.addEdge(succbb, newbb);
			}

			
			for(InstructionHandle h : bytecodes) {
				newbb.addBytecode(h);
			}			
			
			bbGraph.removeNode(branchingBB);
			info.getBasicBlocks().remove(branchingBB);
			info.getBasicBlocks().add(newbb);
			LinkedList<AbstractBasicBlock> exitBBs = getExitNodes(info);
			for(AbstractBasicBlock exitBB : exitBBs) {
				lb.addExitNode(exitBB);
			}
			
			
		}
	}
	

	
	private LinkedList<AbstractBasicBlock> getExitNodes(LoopInfo loopinfo) {
		LinkedList<AbstractBasicBlock> exitBBs = new LinkedList<AbstractBasicBlock>();
		LinkedList<AbstractBasicBlock> loopBB = loopinfo.getBasicBlocks();
		for(AbstractBasicBlock bb : loopBB) {
			for(AbstractBasicBlock succBB : bbGraph.getSuccs(bb)) {
				if(!loopBB.contains(succBB)) {
					if(!exitBBs.contains(bb)){ //&& !bb.equals(loopinfo.getLoopHeader(bbGraph))) {
						exitBBs.add(bb);
					}
				}
			}
		}
		return exitBBs;
	}
	
	private HashMap<Integer, String> getSourceLineMap(JavaClass clazz) {
		String classname = clazz.getClassName();
		//String packageName = "".equals(clazz.getPackageName()) ? "" : clazz.getPackageName() + File.separator;
		String sourceFile;

		ClassFile classFile = ClassLookup.getClassFile(classname);
		sourceFile = classFile.getPath().replace(File.separator + "bin" + File.separator, File.separator + "src" + File.separator).replace(".class", ".java");

		HashMap<Integer, String> lineMap = new HashMap<Integer, String>();
		FileReader fileReader = null;
		try {
			fileReader = new FileReader(sourceFile);
			BufferedReader r = new BufferedReader(fileReader);
			int i = 0;
			String line;
			while ((line = r.readLine()) != null) {
				lineMap.put(i++, line);
			}
		} catch (FileNotFoundException e) {
			System.err.println("Unable to find source code for class " + classname + " - filename: " + sourceFile);
		} catch (IOException e) {
			System.err.println("Error while reading source code for class " + classname + " - filename: " + sourceFile);
		} finally {
			if(fileReader != null) {
				try {
					fileReader.close();
				} catch (IOException e) {
					System.err.println("Could not close file!");
					e.printStackTrace();
				}
			}
		}
		
		return lineMap;
	}
	
	
	private void setLoopBound(LoopInfo info) {
		AbstractBasicBlock loopHeader = info.getLoopHeader(bbGraph);
		LineNumberTable tab = this.currentMethod.getLineNumberTable();
		
		int sourcelinenumber = tab.getSourceLine(loopHeader.getBytecodes().get(0).getPosition());
		int sourceLoopBound = getLoopBound(this.sourceLineMap, sourcelinenumber);
		info.setLoopBound(sourceLoopBound);		
	}
	
	private int getLoopBound(HashMap<Integer, String> lineMap, int sourceLineNumber) {
		int loopBound = -1;
		if(lineMap.isEmpty())
			return loopBound;
		for(int i = sourceLineNumber -1; i > 0; i--) {
			String sourceLine = lineMap.get(Integer.valueOf(i));
			if(sourceLine.indexOf("//@loopbound") != -1) {
				Matcher pMatcher = loopBoundPattern.matcher(sourceLine);
				while(pMatcher.find()) {
					loopBound = Integer.parseInt(pMatcher.group());
					return loopBound;
				}
			}
		}
		if(loopBound == -1) {
			System.err.println("Did not find loopbound in class: " + currentClass.getClassName() + 
								" in method: " + currentMethod.getName() + " on line: " + sourceLineNumber);
			System.err.println("Supported syntax is: //@loopbound = <bound>, where bound is a number");
		}
		
		return loopBound;
	}
	
	private void depthFirstOrdering(AbstractBasicBlock bb) {
		visitedBBs.add(bb);
		for(AbstractBasicBlock succBB : bbGraph.getSuccs(bb)) {
			if(!visitedBBs.contains(succBB))
				depthFirstOrdering(succBB);
		}
		this.bbToDfstValMap.put(bb, this.globalBasicBlockCount);
		--this.globalBasicBlockCount;
	}
	
	private void searchForBackEdges(AbstractBasicBlock bb) {
		visitedBBs.add(bb);
		for(AbstractBasicBlock succBB : bbGraph.getSuccs(bb)) {
			if(this.bbToDfstValMap.get(bb) >= this.bbToDfstValMap.get(succBB)) {
				BackEdge b = new BackEdge(bb, succBB);
				if(!this.backEdges.contains(b)) {
					this.backEdges.add(b);
				}
			}
			if(!visitedBBs.contains(succBB))
				searchForBackEdges(succBB);
		}
	}
	
	private void constructLoopSet(BackEdge backEdge, LinkedList<LoopInfo> foundLoops) {
		AbstractBasicBlock loopHeader = backEdge.getDest();
		
		this.visitedBBs.add(loopHeader);
		
		LoopInfo loopInfo = this.loopHeadersToInfoMap.get(loopHeader);
		
		if(loopInfo != null) {
			if (loopInfo.getBackEdges().contains(backEdge) == false) {
				loopInfo.addBackEdge(backEdge);
			}
		} else {
		
			loopInfo = new LoopInfo();
			loopInfo.addBackEdge(backEdge);
			
			foundLoops.add(loopInfo);
			this.loopHeadersToInfoMap.put(loopHeader, loopInfo);
			loopInfo.addBasicBlock(loopHeader);
		}
		conductReverseDFS(backEdge.getSrc(), loopInfo);		
	}
	
	private void conductReverseDFS(AbstractBasicBlock bb, LoopInfo loopInfo) {
		if (!this.visitedBBs.contains(bb)) {
			this.visitedBBs.add(bb);
			loopInfo.addBasicBlock(bb);
			
			for(AbstractBasicBlock predBB : this.bbGraph.getPreds(bb)) {
				conductReverseDFS(predBB, loopInfo);
			}
		}
	}

}
