package tetasarts.tir;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import tetasarts.exceptions.IncompatibleMergeException;
import tetasarts.exceptions.NodeNotFoundException;
import tetasarts.tir.basicblocks.AbstractBasicBlock;


public class Graph<NodeType extends Node<VisitorType>, VisitorType extends INodeVisitor> implements INodeSubject<VisitorType>{
	private Map<NodeType, List<NodeType>> predMap = new HashMap<NodeType, List<NodeType>>();
	private Map<NodeType, List<NodeType>> succMap = new HashMap<NodeType, List<NodeType>>();
	private List<NodeType> nodeList = new ArrayList<NodeType>();
	Logger logger;
	{
		logger = Logger.getLogger("graph");
		logger.setLevel(Level.OFF);
	}
	
	public Graph(){

	}
	
	public boolean contains(NodeType node){
		return nodeList.contains(node);
	}
	
//	public NodeType getNode(NodeIdentifier ni) {
//		if(!contains(ni)){
//			System.err.println("Graph.getNode()");
//			System.err.println("node not found");
//			return null;
//		} else {
//			return identifierToNodeMap.get(ni);
//		}
//	}
	
	public NodeType merge(NodeType a, NodeType b) throws NodeNotFoundException, IncompatibleMergeException{
		if(!contains(a) || !contains(b))
			throw new NodeNotFoundException("node a "+(contains(a)?"not found":"found" + ", node b ")+
											(contains(b)?"not found":"found"));
		
		logger.log(Level.INFO, "Merging [" + a.getID() + "] and [" + b.getID() + "] into [[");
		logger.log(Level.INFO, "Merging [" + ((AbstractBasicBlock)a).getBytecodes().get(0) + "] and [" + ((AbstractBasicBlock)b).getBytecodes().get(0) + "] into [[");
		boolean addCycle = false;
		NodeType c = (NodeType)a.merge(b);
		logger.log(Level.INFO, c.getID()+"]]");
		logger.log(Level.INFO, ((AbstractBasicBlock)c).getBytecodes().get(0)+"]] " + ((AbstractBasicBlock)c).getBytecodes().size());
		List<NodeType> csucc = new ArrayList<NodeType>();
		csucc.addAll(succMap.get(a));
		addCycle |= csucc.remove(a);
		csucc.remove(b);
		csucc.addAll(succMap.get(b));
		csucc.remove(a);
		addCycle |= csucc.remove(b);
		
		List<NodeType> cpred = new ArrayList<NodeType>();
		cpred.addAll(predMap.get(a));
		addCycle |= cpred.remove(a);
		cpred.remove(b);
		cpred.addAll(predMap.get(b));
		addCycle |= cpred.remove(b);		
		cpred.remove(a);
		
		for(NodeType succ : csucc){
			boolean addC = false;
			addC |= predMap.get(succ).remove(a);
			addC |= predMap.get(succ).remove(b);
			if(addC){
				predMap.get(succ).add(c);
			}			
		}
		for(NodeType pred : cpred){
			boolean addC = false;
			addC |= succMap.get(pred).remove(a);
			addC |= succMap.get(pred).remove(b);
			if(addC){
				succMap.get(pred).add(c);
			}
		}
		
		if(addCycle){
			cpred.add(c);
			csucc.add(c);
		}
		
		a.informRemove(this);
		b.informRemove(this);
		
		succMap.remove(a);
		succMap.remove(b);
		predMap.remove(a);
		predMap.remove(b);
		nodeList.remove(a);
		nodeList.remove(b);
		
		assert nodeList.contains(a) == false : "A is still inside nodelist after merge";
		assert nodeList.contains(b) == false : "A is still inside nodelist after merge";
//		addNode(c);
		predMap.put(c, cpred);
		succMap.put(c, csucc);
		assert getSuccs(c)!=null : "No successor for c after merge";
		assert getPreds(c)!=null : "No predesessor for c after merge";
		nodeList.add(c);
		return c;
	}

	public boolean addNode(NodeType node) {
		assert node!= null : "addnode, trying to add null as node";
		if(contains(node)){
			System.err.println("Graph.addNode()");
			System.err.println("Identifier already in graph");
			return false;
		}
		nodeList.add(node);
		predMap.put(node, new ArrayList<NodeType>());
		succMap.put(node, new ArrayList<NodeType>());
		return true;
	}
	
	public boolean addEdge(NodeType source, NodeType destination) {
		if(contains(source) && contains(destination)){
			succMap.get(source).add(destination);
			predMap.get(destination).add(source);
			return true;
		} else {
			logger.log(Level.WARNING, "Graph.addEdge()");
			logger.log(Level.WARNING, "Node not found");
			return false;
		}
	}
	
	public void removeNode(NodeType node) {
		assert node != null : "node is null!";
		
		
		for(NodeType succ : succMap.get(node)) {
			Iterator<NodeType> iter = predMap.get(succ).iterator();
			while(iter.hasNext()) {
				NodeType nxt = iter.next();
				if(nxt.equals(node)) {
					iter.remove();
				}
			}
		}
		for(NodeType pred : predMap.get(node)) {
			Iterator<NodeType> iter = succMap.get(pred).iterator();
			while(iter.hasNext()) {
				NodeType nxt = iter.next();
				if(nxt.equals(node)) {
					iter.remove();
				}
			}
		}
		succMap.remove(node);
		predMap.remove(node);
		nodeList.remove(node);
	}
	
//	public boolean removeEdge(NodeType source, NodeType destination) {
//		if(contains(source) && contains(destination)){
//			if(!(succMap.get(source).contains(destination) && predMap.get(destination).contains(source))){
//				System.err.println("Graph.removeEdge()");
//				System.err.println("Edge not found.");
//				return false;
//			} else {
//				succMap.get(source).remove(destination);
//				predMap.get(destination).remove(source);
//				return true;
//			}
//		} else {
//			System.err.println("Graph.removeEdge()");
//			System.err.println("Node not found.");
//			return false;
//		}
//	}
	
	public boolean hasSuccessors(NodeType node){
		return succMap.get(node) != null;
	}
	
	public List<NodeType> getSuccs(NodeType node){
		assert node!=null : "Node is null inside getSuccs";
//		int hashCode = node.hashCode();
		assert succMap.get(node) != null : "Successor is NULL!";
		
		List<NodeType> nodeTypeList = succMap.get(node);
		assert nodeTypeList != null : "Node nodeTypeList is null!";
		return Collections.unmodifiableList(nodeTypeList);
	}
	public List<NodeType> getPreds(NodeType node){
		assert node!=null : "Node is null inside getPreds";
		assert predMap.get(node) != null : "Pred is NULL!";
		return Collections.unmodifiableList(predMap.get(node));
	}
	
	public List<NodeType> getNodes(){
		return nodeList;
	}

	@Override
	public void accept(VisitorType v) {
		for(NodeType node : nodeList){
			node.accept(v);
		}
	}	
}
