package tetasarts.uppaal;


import java.util.*;
import org.jdom.*;

import tetasarts.Options;
import tetasarts.Options.PlatformType;
import tetasarts.generator.Constants;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.tir.sets.TemplateSet;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Declarations;
import tetasarts.uppaal.declarations.Int;

public class UppaalTemplate implements Comparable<UppaalTemplate>{
	// template (name, parameter?, declaration?, location*, init?, transition*)
	// name
	private Name name;

	// Parameter?
	private Parameter parameter;

	// declaration?
	private Declarations declaration;

	// location*
	private ArrayList<Location> locations;

	// init?
	private Location init;

	// transition*
	private ArrayList<Transition> transitions;
	
	private MethodInformation methodInformation;
	private int methodSizeInWords;

	public UppaalTemplate(String name){
		this.name = new Name(name);
		this.parameter = null;
		this.declaration = new Declarations();
		this.locations = new ArrayList<Location>();
		this.transitions = new ArrayList<Transition>();
		this.init = null;
//		TemplateSet.globalSet.add(this);
	}
	
	/**
	 * Creates a new empty template with a name.
	 * @param name The name of the template
	 */
	public UppaalTemplate(MethodInformation methodInformation) {
		this(methodInformation, "", new Declarations(), new ArrayList<Location>(), null,
				new ArrayList<Transition>());
	}

	/**
	 * Creates a new template with all properties set
	 * @param name The name of the template
	 * @param parameter The properties of the template
	 * @param declaration Local declarations for the template
	 * @param locations A list of locations in the template
	 * @param init The initial location in the template
	 * @param transitions A list of transitions between locations
	 */
	public UppaalTemplate(MethodInformation methodInformation, String parameter, Declarations declaration,
			ArrayList<Location> locations, Location init,
			ArrayList<Transition> transitions) {
//		this.name = new Name(UppaalModel.convertString("Template_", name));
		this.name = new Name(UppaalModel.getTemplateIdentifier(methodInformation));
		this.methodInformation = methodInformation;
//		methodInformation.setTemplate(this);
		this.parameter = new Parameter(parameter);
		this.declaration = declaration;
		this.declaration.addString("// "+methodInformation.getCanonicalName()+"\n");
		this.locations = locations;
		this.init = init;
		this.transitions = transitions;
		methodSizeInWords = (methodInformation.getBytecodeSize() + 3) / 4;
		//this.declaration.addInteger(new Int("hit", SimpleWcetVisitor.calculateI(true, methodSizeInWords), true));
		//this.declaration.addInteger(new Int("miss", SimpleWcetVisitor.calculateI(false, methodSizeInWords), true));
		if(!Options.EXPLICIT_MODEL) {
			if(Options.PLATFORM_TYPE == PlatformType.JOP) {
				this.declaration.addInteger(new Int("methodSwitchCost", 0));
			}
		}

		TemplateSet.globalSet.add(this);
	}
	
	public void addBacktrack(Location location, String guard) {
		Location backtrack = new Location("backtrack", LocationType.COMMITED, 0, 0);
		Transition backtrackLoop = new Transition(backtrack.getId(), backtrack.getId());
		addLocation(backtrack);
		addTransition(backtrackLoop);

		Transition t = new Transition(location.getId(), backtrack.getId());
		if(!Options.EXPLICIT_MODEL) {
			String wcetInst = "";
			if(Options.PLATFORM_TYPE == PlatformType.JOP){
				if(!Constants.INLINE_WCET) {
					wcetInst = "inst" + location.getId();
					this.getDeclaration().addInteger(new Int(wcetInst, location.getWCET(), true));
				} else {
					wcetInst = Integer.toString(location.getWCET());
				}
			} else if(Options.PLATFORM_TYPE == PlatformType.HVMAVR) {
				//Not sure if this is correct
				wcetInst = "0";
			}
			t.addGuard("executionTime == " + wcetInst);
			t.addGuard(guard);
			t.addUpdate("executionTime = 0");
		}
		addTransition(t);
	}

	public Name getName() {
		return name;
	}

	public void setName(String name) {
		this.name.setName(name);
	}

	public Parameter getParameter() {
		return parameter;
	}

	public void setParameter(String parameter) {
		this.parameter = new Parameter(parameter);
	}

	public Declarations getDeclaration() {
		return declaration;
	}

	public void setDeclaration(Declarations declaration) {
		this.declaration = declaration;
	}

	public ArrayList<Location> getLocations() {
		return locations;
	}

	public void setLocations(ArrayList<Location> locations) {
		this.locations = locations;
	}

	public Location getInit() {
		return init;
	}

	// TODO: init must be in the set of locations
	public void setInit(Location init) {
		if(methodInformation != null)
			init.setComment(methodInformation.getCanonicalName());
		this.init = init;
	}

	public ArrayList<Transition> getTransitions() {
		return transitions;
	}

	public void setTransitions(ArrayList<Transition> transitions) {
		this.transitions = transitions;
	}

	public void addLocation(Location location) {
		this.locations.add(location);
	}

	public void addTransition(Transition transition) {
		this.transitions.add(transition);
	}
	
	public void removeTransition(Location source, Location target) {
		for (int i = 0; i < transitions.size(); i++) {
			if (source.getId() == transitions.get(i).getSource()
					&& target.getId() == transitions.get(i).getTarget()) {
				transitions.remove(i);
				break;
			}
		}
	}
	
	public List<Transition> getTransitions(int sourceID, int targetID){
		ArrayList<Transition> trans = new ArrayList<Transition>();
		for (Transition transition : transitions) {
			if (transition.getSource() == sourceID && transition.getTarget() ==targetID)
				trans.add(transition);
		}
		return trans;
	}
	
	public List<Transition> getIncommingTransitions(Location target) {
		List<Transition> result = new LinkedList<Transition>();
		for(Transition t : getTransitions()){
			if(t.getTarget() == target.getId()) {
				result.add(t);
			}
		}
		return result;
	}
	
	public List<Transition> getOutgoingTransitions(Location source) {
		List<Transition> result = new LinkedList<Transition>();
		for(Transition t : getTransitions()){
			if(t.getSource() == source.getId()) {
				result.add(t);
			}
		}
		return result;
	}
	
	public List<Location> getSuccs(Location head){
		List<Location> result = new LinkedList<Location>();
		for(Transition t : getTransitions()){
			for(Location l : getLocations()){
				if(l.getId() == t.getTarget() && head.getId()==t.getSource()){
					result.add(l);
				}
			}
		}
		return result;
	}
	
	public List<Location> getPreds(Location child){
		List<Location> result = new LinkedList<Location>();
		for(Transition t : getTransitions()){
			for(Location l : getLocations()){
				if(l.getId() == t.getSource() && child.getId()==t.getTarget()){
					result.add(l);
				}
			}
		}
		return result;
	}
	public boolean isBranching(Location l){
		if(getSuccs(l).size()==2){
			return true;
		}
		return false;
	}
	
	public boolean isMerging(Location l){
		if(getPreds(l).size()==2){
			return true;
		}
		return false;
	}
	
	public Location getLeftBranch(Location l){
		return getSuccs(l).get(0);
	}
	public Location getRightBranch(Location l){
		return getSuccs(l).get(1);
	}


	/**
	 * Creates an XML Element object corresponding to the Template object
	 * @return XML Element
	 */
	public Element generateXMLElement() {
		Element result = new Element("template");

		result.addContent(this.name.generateXMLElement());

		if(parameter != null)
			result.addContent(this.parameter.generateXMLElement());

		Element declarationElement = new Element("declaration");
		declarationElement.addContent(this.declaration.toString());
		result.addContent(declarationElement);

		for (Location location : locations) {
			result.addContent(location.generateXMLElement());
		}

		//TODO: make check another place
		if (this.init != null) {
			Element initElement = new Element("init");
			initElement.setAttribute("ref", "id"+this.init.getId());
			result.addContent(initElement);
		}

		for (Transition transition : transitions) {
			result.addContent(transition.generateXMLElement());
		}

		return result;
	}

	@Override
	public int compareTo(UppaalTemplate arg0) {
		
		return this.getName().getName().compareTo(arg0.getName().getName());
	}

	public MethodInformation getMethodInformation() {
		return methodInformation;
	}

	public int getMethodSizeInWords() {
		return methodSizeInWords;
	}
}
