package tetasarts.uppaal.modelcombiner.optimization;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.bcel.generic.Instruction;

import tetasarts.uppaal.modelcombiner.optimization.sync.AutomatonDependencyGraph;
import tetasarts.uppaal.modelcombiner.optimization.sync.AutomatonNode;
import tetasarts.uppaal.modelcombiner.optimization.sync.DependencyEdge;
import uppaal.NTA;

public class ModelCleanerOptimization implements TetaSARTSOptimization {

	private final HashSet<String> jbcOpcodes;
	private AutomatonDependencyGraph depGraph;
	private NTA ntaSystem;
	public ModelCleanerOptimization(AutomatonDependencyGraph depGraph, HashSet<String> jbcOpcodes) {
		this.jbcOpcodes = jbcOpcodes;
		this.depGraph = depGraph;
	}

	@Override
	public void conductOptimization(NTA ntaSystem) {
		this.ntaSystem = ntaSystem;
		removeSuperfluousJBCTemplates(this.jbcOpcodes, ntaSystem);

		removeIsolatedTemplates(this.depGraph, ntaSystem);
	}
	
	private void removeSuperfluousJBCTemplates(HashSet<String> jbcOpcodes, NTA ntaSystem) {

		for(Instruction inst : org.apache.bcel.generic.InstructionConstants.INSTRUCTIONS) {
			if(inst != null) {
				String mnemonic = inst.getName();
				if(!this.jbcOpcodes.contains(mnemonic.toUpperCase())) {
					String opcode = mnemonic.toUpperCase();
					String automatonName = "handle" + opcode + "_OPCODE_T";
					
					removeAutomaton(automatonName);
					removeSystemDeclaration(automatonName);
					removeChannels(automatonName);
				} else {
					System.out.println("Not deleting: " + mnemonic);
				}
			}
		}
	}
	
	private void removeIsolatedTemplates(AutomatonDependencyGraph graph, NTA ntaSys) {
		for(AutomatonNode autNode : graph.getAutomata()) {
			for(DependencyEdge depEdge : autNode.getoutgoingEdges()) {
				if(depEdge.getInitSyncTransition() != null) {
					if(depEdge.getInitSyncTransition().getSync() != null) {
						if(depEdge.getInitSyncTransition().getSync().getChannelName().contains("return_")) {
							break;
						}
					}
				}
			}
			recursivelyRemoveIsolatedTemplates(autNode);
		}
	}
	
	private void recursivelyRemoveIsolatedTemplates(AutomatonNode autNode) {
		for(DependencyEdge edge : autNode.getoutgoingEdges()) {
			if(edge.getInitSyncTransition() != null) {
				if(edge.getInitSyncTransition().getSync() != null) {
					if(edge.getInitSyncTransition().getSync().getChannelName().contains("invoke_")) {
						recursivelyRemoveIsolatedTemplates(edge.getDestination());
					}
				}
			}
		}
		if(!autNode.equals(this.depGraph.getRootNode()) && !autNode.getAutomaton().getName().getName().contains("hwinterface")){
			String automatonName = autNode.getAutomaton().getName().getName();
			removeAutomaton(automatonName);
			System.out.println("Removing: " + automatonName);
			removeSystemDeclaration(automatonName);
			
			removeChannels(automatonName);
		}
	}
	
	private void removeAutomaton(String automatonName) {
		this.ntaSystem.removeAutomaton(automatonName);
	}
	
	private void removeSystemDeclaration(String automatonName) {
		// remove system declaration for removed automaton
		Iterator<String> declIter =  this.ntaSystem.getSystemDeclaration().getSystemInstances().iterator();
		while(declIter.hasNext())
		{
			String decl = declIter.next();
			if(decl.contains(automatonName))
				declIter.remove();
		}
		
		//If the removed element was the last system instance, replace ',' with ';'
		/*LinkedList<String> decls = (LinkedList<String>)ntaSystem.getSystemDeclaration().getDeclarations();
		String lastEle = (String) decls.getLast();
		if(lastEle.contains(",")) {
			lastEle = lastEle.replace(',', ';');		
			decls.removeLast();
			decls.addLast(lastEle);
		}*/
	}
	
	private void removeChannels(String automatonName) {
		// remove declarations of urgent channels for removed automaton
		List<String> ntaDecls = ntaSystem.getDeclarations().getStrings();
		Iterator<String> iter = ntaDecls.iterator();
		while (iter.hasNext()) {
			String iterStr = iter.next();
		    if(iterStr.contains("urgent chan") && iterStr.contains(automatonName.replace("_T", ""))) {
		    	iter.remove();
		    }
		}
	}
}
