package tetasarts.uppaal.modelcombiner.optimization.sync;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import tetasarts.uppaal.modelcombiner.optimization.TetaSARTSOptimization;
import uppaal.Automaton;
import uppaal.Location;
import uppaal.NTA;
import uppaal.Transition;
import uppaal.Location.LocationType;
import uppaal.labels.Guard;
import uppaal.labels.Synchronization;
import uppaal.labels.Synchronization.SyncType;
import uppaal.labels.Update;


public class SynchronizationResolver implements TetaSARTSOptimization {

	private HashSet<AutomatonNode> visitedNodes;
	private NTA ntasys;

	private AutomatonDependencyGraph graph;
	
	public SynchronizationResolver(AutomatonDependencyGraph depGraph) {
		this.graph = depGraph;
	}
	
	@Override
	public void conductOptimization(NTA ntaSystem) {		
		this.visitedNodes = new HashSet<AutomatonNode>();
		this.ntasys = ntaSystem;
		AutomatonNode root = graph.getRootNode();
		traverseAndResolveSynchronizations(root);
		
		removeVistedAutomata(this.visitedNodes);
	}
	
	private void removeVistedAutomata(HashSet<AutomatonNode> visitedNodes) {	
		
		for(AutomatonNode node : visitedNodes) {
			if(!node.getAutomaton().getName().getName().equals(this.graph.getRootNode().getAutomaton().getName().getName())) {
				this.ntasys.removeAutomaton(node.getAutomaton());
				
				boolean reachedSystem = false;
				Iterator<String> iter = this.ntasys.getSystemDeclaration().getDeclarations().iterator();
				while(iter.hasNext()) {
					String decl = iter.next();
					if(decl.contains("system"))
						reachedSystem = true;
					
					if(reachedSystem == true) {
						if(decl.contains(node.getAutomaton().getName().getName())) {
							iter.remove();
						}
					}
				}
			}
		}
		String lastDecl = "";
		Iterator<String> iter = this.ntasys.getSystemDeclaration().getDeclarations().iterator();
		while(iter.hasNext()) {
			String decl = iter.next();
			if(!iter.hasNext()) {
				iter.remove();
				lastDecl = decl.replace(',', ';');
			}
		}
		this.ntasys.getSystemDeclaration().addDeclaration(lastDecl);

	}
	
	private void traverseAndResolveSynchronizations(AutomatonNode startNode) {
		recursivelyVisitAndResolve(startNode);
		
	}
	
	private void recursivelyVisitAndResolve(AutomatonNode node) {
		for(DependencyEdge outgoingEdge : node.getoutgoingEdges()) {
			if(outgoingEdge.getInitSyncTransition().getSync() != null) {
				String outgoingChannelName = outgoingEdge.getInitSyncTransition().getSync().getChannelName();
				
				if(outgoingChannelName.startsWith("invoke")) {
					String opcodeNameOfChannel = extractOpcodeNameFromChannel(outgoingChannelName);
					
					AutomatonNode receivingAutomatonNode = outgoingEdge.getDestination();

						for(DependencyEdge receivingAutNodeOutgoingEdge : receivingAutomatonNode.getoutgoingEdges()) {
							if(receivingAutNodeOutgoingEdge.getInitSyncTransition().getSync() != null) {
							String returningChannelName = receivingAutNodeOutgoingEdge.getInitSyncTransition().getSync().getChannelName();
							
								if(opcodeNameOfChannel.equals(returningChannelName) &&
								   receivingAutNodeOutgoingEdge.getDestination().equals(node)) {
									
									recursivelyVisitAndResolve(receivingAutomatonNode);
									copyTemplateToAutomaton(receivingAutomatonNode, node);
									
									Transition originInitTrans = null,
											   originReceiverTrans = null,
											   destInitTrans = null,
											   destReceiverTrans = null;
									
									for(Transition trans : node.getAutomaton().getTransitions()) {
										if(trans.getSync() != null) {
											if(trans.getSync().getSyncType() == SyncType.INITIATOR) {
												if(trans.getSync().getChannelName().equals(outgoingEdge.getInitSyncTransition().getSync().getChannelName()))
													originInitTrans = trans;
												else if(trans.getSync().getChannelName().equals(receivingAutNodeOutgoingEdge.getInitSyncTransition().getSync().getChannelName())) {
													destInitTrans = trans;
												}
											}
											else if(trans.getSync().getSyncType() == SyncType.RECEIVER) {
												if(trans.getSync().getChannelName().equals(receivingAutNodeOutgoingEdge.getRecSyncTransition().getSync().getChannelName()))
													originReceiverTrans = trans;
												else if(trans.getSync().getChannelName().equals(outgoingEdge.getRecSyncTransition().getSync().getChannelName())) {
													destReceiverTrans = trans;
												}
											}
										}
									}
									eliminateSync(node.getAutomaton(), originInitTrans, originReceiverTrans, destInitTrans, destReceiverTrans);
								}
							}
							
						}
					}
				}
			}
			this.visitedNodes.add(node);
	}
	
	private String extractOpcodeNameFromChannel(String channelName) {
		String opcode;
		if(channelName.startsWith("invoke")) {
			opcode = channelName.replace("invoke", "return");
		} else if(channelName.startsWith("return")) {
			opcode = channelName.replace("return", "invoke");
		} else {
			opcode = "";
		}
		return opcode;
	}
	
	private HashMap<Location, Location> oldToNewLoc = new HashMap<Location, Location>();
	
	private static int node_number = 0;
	private void copyTemplateToAutomaton(AutomatonNode fromAutomatonNode, AutomatonNode toAutomatonNode) {
		
		
		Automaton fromAutomaton = fromAutomatonNode.getAutomaton();
		fromAutomaton.setInit(null);
		Automaton toAutomaton = toAutomatonNode.getAutomaton();
		
		for(Location loc : fromAutomaton.getLocations()) {
			
			String newLocName = (loc.getName() == null) ? toAutomaton.getName().getName() + "_" : loc.getName().getName() + "_" + toAutomaton.getName().getName() + "_" + node_number++;
			
			Location newLoc = new Location(toAutomaton, newLocName);
			
			newLoc.setInvariant(loc.getInvariant());
			newLoc.setComment(loc.getComment());
			newLoc.setType(loc.getType());
			newLoc.setPositioned(loc.isPositioned());
			newLoc.setTransitional(loc.isTransitional());
			
			oldToNewLoc.put(loc, newLoc);
		}
		
		for(Transition trans : fromAutomaton.getTransitions()) {
			Location originalSource = trans.getSource();
			Location originalDest = trans.getTarget();
			
			Location newSource = oldToNewLoc.get(originalSource);
			Location newDest = oldToNewLoc.get(originalDest);
			
			/*
			 * Transition is automatically added to the automaton
			 */
			Transition newTrans = new Transition(toAutomaton, newSource, newDest);
			
			newTrans.setControllable(trans.isControllable());
			newTrans.setGuard(new Guard(trans.getGuard()));
			newTrans.setNails(trans.getNails());
			if(trans.getSync() != null)
				newTrans.setSync(new Synchronization(trans.getSync().getChannelName(), trans.getSync().getSyncType()));
			newTrans.setUpdate(new Update(trans.getUpdate()));
		}
		
		List<String> toAutDeclStrs = toAutomaton.getDeclaration().getStrings();
		for(String declStr : fromAutomaton.getDeclaration().getStrings()) {
			if(!toAutDeclStrs.contains(declStr))
				toAutDeclStrs.add(declStr);	
		}
	}
	
	private void eliminateSync(Automaton aut, Transition originInitTrans, Transition originReceiverTrans, Transition destInitTrans, Transition destReceiverTrans) {

		if(originReceiverTrans == null || destInitTrans == null)
			return;
		
		if(originInitTrans.getGuard() != null) {	
			Location origin = originInitTrans.getTarget();
			origin.setType(LocationType.COMMITTED);
			removeChannelDeclaration(this.ntasys, destInitTrans.getSync().getChannelName());
			originInitTrans.setSync(null);

			Location originalOriginReceiverTarget = originReceiverTrans.getTarget();	
			
			
			originReceiverTrans.setTarget(destReceiverTrans.getTarget());
			originReceiverTrans.setSync(null);
			if(destReceiverTrans.getUpdate() != null)
				originReceiverTrans.setUpdate(new Update(destReceiverTrans.getUpdate()));
			
			Location destWaitLocation = destInitTrans.getTarget();
			
			destInitTrans.getSource().setType(LocationType.COMMITTED);
			destInitTrans.setTarget(originalOriginReceiverTarget);
			destInitTrans.setSync(null);
			
			// cleanup
			aut.removeTransition(destReceiverTrans);
			removeChannelDeclaration(this.ntasys, destReceiverTrans.getSync().getChannelName());
			
			aut.removeLocation(destWaitLocation);	
		}
		else {
			Location originWaitLocation = originInitTrans.getTarget();
			Location destWaitLocation = destInitTrans.getTarget();
			originInitTrans.setTarget(destReceiverTrans.getTarget());
			originInitTrans.setSync(null);
			originInitTrans.getSource().setType(LocationType.COMMITTED);
			
			
			destInitTrans.getSource().setType(LocationType.COMMITTED);
			destInitTrans.setTarget(originReceiverTrans.getTarget());
			destInitTrans.setSync(null);
			
			//cleanup
			aut.removeLocation(originWaitLocation);
			aut.removeLocation(destWaitLocation);
			
			aut.removeTransition(destReceiverTrans);
			removeChannelDeclaration(this.ntasys, destReceiverTrans.getSync().getChannelName());
			aut.removeTransition(originReceiverTrans);
			removeChannelDeclaration(this.ntasys, originReceiverTrans.getSync().getChannelName());
		}
	}
	
	private void removeChannelDeclaration(NTA ntaSystem, String channelName) {
		List<String> ntaDecls = ntaSystem.getDeclarations().getStrings();
		Iterator<String> iter = ntaDecls.iterator();
		while (iter.hasNext()) {
			String iterStr = iter.next();
		    if(iterStr.contains("urgent chan") && iterStr.contains(channelName)) {
		    	iter.remove();
		    }
		}
	}
}
