package tetasarts.uppaal.schedulers;


import tetasarts.Options;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Nail;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Channel;
import tetasarts.uppaal.declarations.Declarations;

public class FifoScheduler extends AbstractSchedulerUppaalModel{
	
	@Override
	public SchedulerTemplate getScheduler() {
		SchedulerTemplate scheduler = new SchedulerTemplate("FIFO_scheduler");

		Location initial = super.createLoc("", LocationType.COMMITED, 296, -48);
		scheduler.addLocation(initial);
		scheduler.setInit(initial);
		
		Location runningLoc = super.createLoc("Running", LocationType.NORMAL, 296, -160);
		runningLoc.setInvariant("executionTime <= scheduler_wcet &&\nexecutionTime' == running[schedulerID]", runningLoc.getPosX(), runningLoc.getPosY() - 35);
		scheduler.addLocation(runningLoc);
		
		Transition initRunTrans = new Transition(initial.getId(), runningLoc.getId());
		int syncXPos = (runningLoc.getPosX() + initial.getPosX()) / 2;
		int syncYPos = (runningLoc.getPosY() + initial.getPosY()) / 2;
		initRunTrans.setSync(new Channel("GO", "not_go??", true), true, syncXPos, syncYPos - 15); // check if this is correct
		initRunTrans.setUpdate("init_scheduler(),\nexecutionTime = 0", syncXPos, syncYPos);
		scheduler.addTransition(initRunTrans);
		
		Location exerciseboundLoc = super.createLoc("finishedHyperPeriod", LocationType.COMMITED, 32, -48);
		scheduler.addLocation(exerciseboundLoc);
		Transition exerciseLockTrans = new Transition(exerciseboundLoc.getId(), exerciseboundLoc.getId());
		exerciseLockTrans.addNail(new Nail(2, -78));
		exerciseLockTrans.addNail(new Nail(62, -78));
		scheduler.addTransition(exerciseLockTrans);
		
		Transition exerciseBoundTrans = new Transition(runningLoc.getId(), exerciseboundLoc.getId());
		int exerXPos = (runningLoc.getPosX() + exerciseboundLoc.getPosX()) / 2;
		int exerYPos = (runningLoc.getPosY() + exerciseboundLoc.getPosY()) / 2;
		exerciseBoundTrans.setGuard("globalClock >= exerciseBound", exerXPos - 100, exerYPos);
		scheduler.addTransition(exerciseBoundTrans);
		
		Location scheduleLoc = super.createLoc("schedule", LocationType.COMMITED, 296, -280);
		scheduler.addLocation(scheduleLoc);
		
		Transition runningScheduleTrans = new Transition(runningLoc.getId(), scheduleLoc.getId());
		int runningXPos = (runningLoc.getPosX() + scheduleLoc.getPosX()) / 2;
		int runningYPos = (runningLoc.getPosY() + scheduleLoc.getPosY()) / 2;
		runningScheduleTrans.setGuard("executionTime == scheduler_wcet &&\nglobalClock < exerciseBound", 
				runningXPos - 100, 
				runningYPos);
		scheduler.addTransition(runningScheduleTrans);
		
		Transition notSchedulableTrans = new Transition(scheduleLoc.getId(), runningLoc.getId());
		notSchedulableTrans.addNail(new Nail(32, -280));
		notSchedulableTrans.addNail(new Nail(32, -160));
		int notSchedX = ((scheduleLoc.getPosX() + 32) / 2) - 110;
		int notSchedY = scheduleLoc.getPosY() - 30;
		notSchedulableTrans.setGuard("!exists(i:ThreadID)schedulable[i]", notSchedX, notSchedY - 15);
		notSchedulableTrans.setUpdate("idle(),\nexecutionTime = 0", notSchedX, notSchedY);
		scheduler.addTransition(notSchedulableTrans);
		
		Transition schedulableTrans = new Transition(scheduleLoc.getId(), runningLoc.getId());
		schedulableTrans.addNail(new Nail(568, -280));
		schedulableTrans.addNail(new Nail(568, -160));
		int schedX = ((scheduleLoc.getPosX() + 568) / 2) - 60;
		int schedY = scheduleLoc.getPosY() - 30;		
		schedulableTrans.setGuard("exists(i:ThreadID)schedulable[i]", schedX, schedY - 15);
		schedulableTrans.setUpdate("select_thread(),\nexecutionTime = 0", schedX, schedY);
		scheduler.addTransition(schedulableTrans);
		
		scheduler.setDeclaration(this.getFifoSchedulerDeclarations());
		
		return scheduler;
	}

	

	
	private Declarations getFifoSchedulerDeclarations() {
		Declarations decl = new Declarations();

		StringBuilder sb = new StringBuilder();
		
		sb.append("/*\n")
		.append("* The scheduler_wcet constant needs to be set according to the specific scheduler used\n")
		.append("*/\n\n")
		.append("const int scheduler_wcet = 1;\n\n")
		.append("int selectedThread = -1;\n")
		.append("clock executionTime;\n\n")
		.append("void init_scheduler() {\n")
		.append("    fifo_cursor = 0;\n")
		.append("}\n\n")
		.append("void update_queue() {\n")
		.append("    int i;\n")
		.append("    for(i = 1; i <= fifo_size - 1; i++) {\n")
		.append("        fifo_queue[i-1] = fifo_queue[i];\n")
		.append("    }\n")
		.append("    fifo_cursor--;\n")
		.append("}\n\n")
		.append("void select_task_for_running(const ThreadID threadID) {\n")
		.append("    int i;\n")
		.append("    for (i = 0; i <= totalThreads; i++){\n")
		.append("		running[i] = 0;\n")
		.append("	}\n")
		.append("	running[threadID] = 1;\n")
		.append((Options.UTILIZATION_MONITOR) ? "		idling = false;\n" : "")
		.append("}\n\n")
		.append("void select_thread() {\n")
		.append("    selectedThread = fifo_queue[0].threadID;\n")
		.append("    update_queue();\n")
		.append("    select_task_for_running(selectedThread);\n")
		.append("}\n\n")
		.append("void idle(){\n")
		.append("	int i;\n")
		.append("	for (i = 0; i <= totalThreads; i++){\n")
	  	.append("		running[i] = 0;\n")
	  	.append("	}\n")
	  	.append((Options.UTILIZATION_MONITOR) ? "		idling = true;\n" : "")
	  	.append("}\n");
		
		decl.addString(sb.toString());
		return decl;
	}
	
	@Override
	public PeriodicControllerTemplate getPeriodicController() {
		return super.generatePeriodicControllerTemplate("insert_task_in_scheduler(pID)", "task_done_scheduler()");
	}

	@Override
	public SporadicControllerTemplate getSporadicController() {
		return super.generateSporadicControllerTemplate("insert_task_in_scheduler(sID)", "task_done_scheduler()");
	}

	@Override
	public Declarations getDeclarations() {
		Declarations decl = new Declarations();
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("/*\n")
				.append("* This is for the FIFO-scheduler\n")
				.append("*/\n")
				.append("typedef struct {\n")
				.append("    ThreadID threadID;\n")
				.append("} schedulable_task;\n")
				.append("// Maybe the size of the fifo queue should be larger to accommodate that a periodic task\n")
				.append("// is released again before its initial release has been handled\n")
				.append("const int fifo_size = totalThreads + 1;\n")
				.append("int[0,fifo_size - 1] fifo_cursor;\n")
				.append("schedulable_task fifo_queue[fifo_size];\n")
				.append("void insert_task_in_scheduler(const ThreadID threadID) {\n")
				.append("    int i;\n")
				.append("	if (synchronized){\n")
				.append("		interruptWaiting = true;\n")
				.append("	}\n")
				.append("    fifo_queue[fifo_cursor].threadID = threadID;\n")
				.append("    fifo_cursor++;\n")
				.append("    if(!exists(i:ThreadID) running[i]) {\n")
				.append("        running[schedulerID] = 1;\n")
				.append("    }\n")
				.append("}\n")
				.append("void task_done_scheduler() {\n")
				.append("    int i;\n")
				.append("	if (synchronized){\n")
				.append("		interruptWaiting = true;\n")
				.append("	} else {\n")
				.append("		for (i = 0; i <= totalThreads; i++){\n")
				.append("			running[i] = 0;\n")
				.append("		}\n")
				.append("		running[schedulerID] = 1;\n")
				.append("	}\n")
				.append("}\n")
				.append("/*\n")
				.append("* End of FIFO-scheduler\n")
				.append("*/");
		decl.addString(sb.toString());
		return decl;
	}
}
