package tetasarts.uppaal.schedulers;

import tetasarts.Options;
import tetasarts.generator.Constants;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Nail;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Channel;
import tetasarts.uppaal.declarations.Declarations;

public class FpsScheduler extends AbstractSchedulerUppaalModel {

	@Override
	public SchedulerTemplate getScheduler() {
		SchedulerTemplate scheduler = new SchedulerTemplate("FPS_scheduler");

		Location initial = super.createLoc("", LocationType.COMMITED, 296, -48);
		scheduler.addLocation(initial);
		scheduler.setInit(initial);
		
		Location runningLoc = super.createLoc("Running", LocationType.NORMAL, 296, -160);
		runningLoc.setInvariant("executionTime <= scheduler_wcet &&\nexecutionTime' == running[schedulerID]", runningLoc.getPosX(), runningLoc.getPosY() - 35);
		if(Options.STATISTICAL_MODEL_CHECKING) {
			runningLoc.setExponentialRate(Constants.EXPONENTIAL_RATE_STR);
		}
		scheduler.addLocation(runningLoc);
		
		Transition initRunTrans = new Transition(initial.getId(), runningLoc.getId());
		int syncXPos = (runningLoc.getPosX() + initial.getPosX()) / 2;
		int syncYPos = (runningLoc.getPosY() + initial.getPosY()) / 2;
		initRunTrans.setSync(new Channel("GO", "not_go??", true), true, syncXPos, syncYPos - 15); // check if this is correct
		initRunTrans.setUpdate("executionTime = 0", syncXPos, syncYPos);
		scheduler.addTransition(initRunTrans);
		
		Location exerciseboundLoc = super.createLoc("finishedHyperPeriod", LocationType.COMMITED, 32, -48);
		scheduler.addLocation(exerciseboundLoc);
		Transition exerciseLockTrans = new Transition(exerciseboundLoc.getId(), exerciseboundLoc.getId());
		exerciseLockTrans.addNail(new Nail(2, -78));
		exerciseLockTrans.addNail(new Nail(62, -78));
		scheduler.addTransition(exerciseLockTrans);
		
		Transition exerciseBoundTrans = new Transition(runningLoc.getId(), exerciseboundLoc.getId());
		int exerXPos = (runningLoc.getPosX() + exerciseboundLoc.getPosX()) / 2;
		int exerYPos = (runningLoc.getPosY() + exerciseboundLoc.getPosY()) / 2;
		exerciseBoundTrans.setGuard("globalClock >= exerciseBound", exerXPos - 100, exerYPos);
		scheduler.addTransition(exerciseBoundTrans);
		
		Location scheduleLoc = super.createLoc("schedule", LocationType.COMMITED, 296, -280);
		scheduler.addLocation(scheduleLoc);
		
		Transition runningScheduleTrans = new Transition(runningLoc.getId(), scheduleLoc.getId());
		int runningXPos = (runningLoc.getPosX() + scheduleLoc.getPosX()) / 2;
		int runningYPos = (runningLoc.getPosY() + scheduleLoc.getPosY()) / 2;
		runningScheduleTrans.setGuard("executionTime == scheduler_wcet &&\nglobalClock < exerciseBound", 
				runningXPos - 100, 
				runningYPos);
		scheduler.addTransition(runningScheduleTrans);
		
		Transition notSchedulableTrans = new Transition(scheduleLoc.getId(), runningLoc.getId());
		notSchedulableTrans.addNail(new Nail(32, -280));
		notSchedulableTrans.addNail(new Nail(32, -160));
		int notSchedX = ((scheduleLoc.getPosX() + 32) / 2) - 110;
		int notSchedY = scheduleLoc.getPosY() - 30;
		notSchedulableTrans.setGuard("!exists(i:ThreadID)schedulable[i]", notSchedX, notSchedY - 15);
		notSchedulableTrans.setUpdate("idle(),\nexecutionTime = 0", notSchedX, notSchedY);
		scheduler.addTransition(notSchedulableTrans);
		
		Transition schedulableTrans = new Transition(scheduleLoc.getId(), runningLoc.getId());
		schedulableTrans.addNail(new Nail(568, -280));
		schedulableTrans.addNail(new Nail(568, -160));
		int schedX = ((scheduleLoc.getPosX() + 568) / 2) - 60;
		int schedY = scheduleLoc.getPosY() - 30;		
		schedulableTrans.setGuard("exists(i:ThreadID)schedulable[i]", schedX, schedY - 15);
		schedulableTrans.setUpdate("select_thread(),\nexecutionTime = 0", schedX, schedY);
		scheduler.addTransition(schedulableTrans);
		
		scheduler.setDeclaration(this.getFPSSchedulerDeclarations());
		
		return scheduler;
	}
	
	private Declarations getFPSSchedulerDeclarations() {
		Declarations decl = new Declarations();

		StringBuilder sb = new StringBuilder();
		
		sb.append("/*\n")
			  .append("* The scheduler_wcet constant needs to be set according to the specific scheduler used\n")
			  .append("*/\n")
			  .append("const int scheduler_wcet = 1;\n")
			  .append("int selectedThread;\n")
			  .append("clock executionTime;\n\n")
			  .append("void select_thread(){\n")
			  .append("		int i;\n")
			  .append("		selectedThread = -1;\n")
			  .append("		selectedThreadPriority = -1;\n")
			  .append("		for(i:ThreadID){\n")
			  .append("			if(schedulable[i] && threadPriority[i] > selectedThreadPriority){\n")
			  .append("				selectedThread = i;\n")
			  .append("				selectedThreadPriority = threadPriority[i];\n")
			  .append("			}\n")
			  .append("		}\n\n")
			  .append("		for (i = 0; i <= totalThreads; i++){\n")
			  .append("			running[i] = 0;\n")
			  .append("		}\n")
			  .append("		running[selectedThread] = 1;\n")
			  .append((Options.UTILIZATION_MONITOR) ? "		idling = false;\n" : "")
			  .append("}\n\n")
			  .append("void idle(){\n")
			  .append("		int i;\n")
			  .append("		for (i = 0; i <= totalThreads; i++){\n")
			  .append("			running[i] = 0;\n")
			  .append("		}\n")
			  .append("		selectedThreadPriority = -1;\n")
			  .append((Options.UTILIZATION_MONITOR) ? "		idling = true;\n" : "")
			  .append("}\n");
		
		decl.addString(sb.toString());
		return decl;
	}
	

	@Override
	public PeriodicControllerTemplate getPeriodicController() {
		return super.generatePeriodicControllerTemplate("enqueue_task(pID, threadPriority[pID])", "dequeue_task(pID)");
	}

	@Override
	public SporadicControllerTemplate getSporadicController() {
		return super.generateSporadicControllerTemplate("enqueue_task(sID, threadPriority[sID])", "dequeue_task(sID)");
	}

	@Override
	public Declarations getDeclarations() {
		Declarations decl = new Declarations();
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("/*\n")
				.append("* This is for the FPS-scheduler\n")
				.append("*/\n")
				.append("int selectedThreadPriority = -1;\n\n")
				.append("void enqueue_task(ThreadID tid, const int pri) {\n")
				.append("	if(pri < selectedThreadPriority) {\n")
				.append("		return;\n")
				.append("	} else if(pri >= selectedThreadPriority) {\n")
	        	.append("		int i;\n")
		    	.append("		if(synchronized){\n")
				.append("		    blocked[tid] = true;\n")
				.append("	    } else {\n")
				.append("	    	for(i = 0; i <= totalThreads; i++){\n")
				.append("		    	running[i] = 0;\n")
				.append("	       	}\n")
				.append("	        running[schedulerID] = 1;\n")
				.append("		}\n")
				.append("	}\n")
				.append("}\n\n")
				.append("void dequeue_task(ThreadID tid) {\n")
				.append("	int i;\n")
				.append("	if(synchronized){\n")
				.append("		    blocked[tid] = true;\n")
				.append("	} else {\n")
				.append("		for(i = 0; i <= totalThreads; i++){\n")
				.append("			running[i] = 0;\n")
				.append("		}\n")
				.append("		running[schedulerID] = 1;\n")
				.append("	}\n")
				.append("}\n")
				.append("/*\n")
				.append("* End of FPS-scheduler\n")
				.append("*/");
		decl.addString(sb.toString());
		return decl;
	}
}
