package tetasarts.uppaal.translation;

import java.util.Hashtable;
import java.util.Set;

import tetasarts.Options;
import tetasarts.Options.PlatformType;
import tetasarts.analysis.SimpleWcetVisitor;
import tetasarts.generator.Constants;
import tetasarts.platforms.HVMAVR;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.classinfo.AbstractClass;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.classinfo.IClassVisitor;
import tetasarts.tir.classinfo.NormalClass;
import tetasarts.tir.classinfo.PeriodicThread;
import tetasarts.tir.classinfo.SporadicTask;
import tetasarts.tir.classinfo.SporadicThread;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.tir.sets.MethodSet;
import tetasarts.tir.sets.TemplateSet;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.PeriodicRunMethod;
import tetasarts.uppaal.PublicMethod;
import tetasarts.uppaal.SporadicRunMethod;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.UppaalTemplate;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Channel;
import tetasarts.uppaal.declarations.Clock;
import tetasarts.uppaal.translation.TransitionCreaterVisitor;

public class TemplateCreatorClassVisitor implements IClassVisitor{
	private UppaalModel uppaalModel;
	private ClassGraph classGraph;
	private Set<SporadicTask> sporadicTasks;
	
	public TemplateCreatorClassVisitor(UppaalModel model, ClassGraph classGraph, Set<SporadicTask> sporadicTasks){
		this.uppaalModel = model;
		this.classGraph = classGraph;
		this.sporadicTasks = sporadicTasks;
	}

	@Override
	public void visitAbstractClass(AbstractClass s) {
		for (MethodInformation methodInfo : s.getMethods().values()) {
			if(methodInfo.isUsed())
				generatePublicMethodTemplate(methodInfo);
		}
	}

	@Override
	public void visitNormalClass(NormalClass s) {
		for (MethodInformation methodInfo : s.getMethods().values()) {
			if(methodInfo.isUsed())
				generatePublicMethodTemplate(methodInfo);
		}
	}

	@Override
	public void visitPeriodicThread(PeriodicThread s) {
		for (MethodInformation methodInfo : s.getMethods().values()) {
			if(methodInfo.isUsed()) {
				if (methodInfo.getName().equals("run") && methodInfo.getSignature().equals("()Z")
						&& !Constants.DISABLE_TASKS){
					PeriodicRunMethod template = new PeriodicRunMethod(methodInfo);
					generateRunMethodTemplate(methodInfo, template);
				} else {
					generatePublicMethodTemplate(methodInfo);
				}
			}
		}
	}

	@Override
	public void visitSporadicThread(SporadicThread s) {
		for (MethodInformation methodInfo : s.getMethods().values()) {
			if(methodInfo.isUsed()) {
				if (methodInfo.getName().equals("run") && methodInfo.getSignature().equals("()Z")
						&& !Constants.DISABLE_TASKS){
					SporadicRunMethod template = new SporadicRunMethod(methodInfo);
					generateRunMethodTemplate(methodInfo, template);
				} else {
					generatePublicMethodTemplate(methodInfo);
				}
			}
		}
	}
	
	private String getValidMethodLabelName(MethodInformation methodInfo){
		return methodInfo.getCanonicalName().replaceAll("\\W++", "_"); 
	}
	
	private void generateRunMethodTemplate(MethodInformation methodInfo, UppaalTemplate template) {
		assert methodInfo.isGenerated() : "RunMethod-methodInfo not generated";
		template.setParameter("const ThreadID tID");
		if(!Options.EXPLICIT_MODEL) {
			template.getDeclaration().addClock(new Clock("executionTime"));
		}
		if (Constants.USE_WCET){
			template.getDeclaration().addClock(new Clock("wcet"));
		}
		Hashtable<AbstractBasicBlock, Location> locationMap = 
		generateMethodBody(template, methodInfo.getInstructionGraph());
		//System.out.println("cname: " + methodInfo.getCanonicalName());
		
		for(AbstractBasicBlock b:  methodInfo.getInstructionGraph().getHeads()){
			System.out.println(b.getName());
		}
		
		assert methodInfo.getInstructionGraph().getTails().size() == 1 : "Nr of tails on graph was not 1, count : " + methodInfo.getInstructionGraph().getTails().size() + " in method: " + methodInfo.getCanonicalName();
		assert methodInfo.getInstructionGraph().getHeads().size() == 1 : "Nr of heads on graph was not 1, count : " + methodInfo.getInstructionGraph().getHeads().size() + " in method: " + methodInfo.getCanonicalName();
		
		Location start = locationMap.get(methodInfo.getInstructionGraph().getHeads().get(0));
		Location end = locationMap.get(methodInfo.getInstructionGraph().getTails().get(0));

		if (start != null && end != null){
			if(Options.EXPLICIT_MODEL) {
				start.setType(LocationType.COMMITED);
				end.setType(LocationType.COMMITED);
			}
			Location initial = new Location(getValidMethodLabelName(methodInfo)+"_init");
			if (Constants.USE_WCET){
				initial.setInvariant("wcet' == running[tID]");
			}
			template.addLocation(initial);
			template.setInit(initial);
			
			Transition initialTransition = new Transition(initial.getId(), start.getId());
			initialTransition.setSyncRunOrFire("run[tID]?");
			if(!Options.EXPLICIT_MODEL) {
				initialTransition.addUpdate("executionTime = 0");
			}
			if (Constants.USE_WCET){
				initialTransition.addUpdate("wcet = 0");
			}
			Transition endTransition = new Transition(end.getId(), initial.getId());
			endTransition.setSyncRunOrFire("run[tID]!");
			if(!Options.EXPLICIT_MODEL) {
				if(Options.PLATFORM_TYPE == PlatformType.JOP) {
					String wcetInst = "inst" + end.getId();
					endTransition.addGuard("executionTime == " + (Constants.INLINE_WCET?end.getWCET():wcetInst));
				} else if(Options.PLATFORM_TYPE == PlatformType.HVMAVR) {
					HVMAVR hvmavrPlatform = HVMAVR.getInstance();
					endTransition.addGuard("executionTime == 0");
				}
				endTransition.addUpdate("executionTime = 0");
			}
			template.addTransition(initialTransition);
			template.addTransition(endTransition);
		}
		
		uppaalModel.addTemplate(template);
	}
	static int c = 0;
	private void generatePublicMethodTemplate(MethodInformation methodInfo){
		assert methodInfo.isGenerated() : "Methodinfo not generated!";
		String channelName = UppaalModel.getChannelIdentifier(methodInfo);
		uppaalModel.getDeclarations().addChannel(new Channel(channelName, methodInfo.getCanonicalName(), "ThreadID", (Options.STATISTICAL_MODEL_CHECKING) ? true : false));
		PublicMethod template = new PublicMethod(methodInfo);
		template.setParameter("const ThreadID tID");
		
		if(!Options.EXPLICIT_MODEL) {
			template.getDeclaration().addClock(new Clock("executionTime"));
		}
		
		if (Constants.USE_WCET){
			template.getDeclaration().addClock(new Clock("wcet"));
		}
		BasicBlockGraph instructionGraph = methodInfo.getInstructionGraph();
		Hashtable<AbstractBasicBlock, Location> locationMap = 
			generateMethodBody(template, instructionGraph);
		
		assert methodInfo.getInstructionGraph().getHeads().size() == 1 : "Nr of heads on graph was not 1, count : " + methodInfo.getInstructionGraph().getHeads().size() + " in method: " + methodInfo.getCanonicalName();;
		assert methodInfo.getInstructionGraph().getTails().size() == 1 : "Nr of tails on graph was not 1, count : " + methodInfo.getInstructionGraph().getTails().size() + " in method: " + methodInfo.getCanonicalName();;
		
		Location start = locationMap.get(methodInfo.getInstructionGraph().getHeads().get(0));
		Location end = locationMap.get(methodInfo.getInstructionGraph().getTails().get(0));
		
		if (start != null && end != null){
			if(Options.EXPLICIT_MODEL) {
				start.setType(LocationType.COMMITED);
				end.setType(LocationType.COMMITED);
			}
			Location initial = new Location(getValidMethodLabelName(methodInfo)+"_init"); // initial
			if (Constants.USE_WCET){
				initial.setInvariant("wcet' == running[tID]");
			}
			if(Options.STATISTICAL_MODEL_CHECKING)
				initial.setExponentialRate(Constants.EXPONENTIAL_RATE_STR);
			template.addLocation(initial);
			template.setInit(initial);
			
			Transition initialTransition = new Transition(initial.getId(), start.getId());
			initialTransition.setSync(new Channel(channelName, methodInfo.getCanonicalName(), "tID"), false);
			if(!Options.EXPLICIT_MODEL) {
				initialTransition.addUpdate("executionTime = 0");
			}
			if (Constants.USE_WCET){
				initialTransition.addUpdate("wcet = 0");
			}
			//TODO use method cache to determine miss or hit
			if(!Options.EXPLICIT_MODEL && Options.PLATFORM_TYPE == PlatformType.JOP) {
				int l = SimpleWcetVisitor.calculateI(false, template.getMethodSizeInWords());
				if (l <= 37)
					l = 0;
				else
					l -= 37;
				initialTransition.addUpdate("methodSwitchCost = " + l);
				if (Constants.USE_WCET){
					start.setInvariant("executionTime <= methodSwitchCost &&\nexecutionTime' == running[tID] &&\nwcet' == running[tID]");
				} else {
					start.setInvariant("executionTime <= methodSwitchCost &&\nexecutionTime' == running[tID]");
				}
				for (Transition transition : template.getOutgoingTransitions(start)) {
					//use set instead of add, because the old value must be overwritten.
					transition.setGuard("executionTime == methodSwitchCost");
					if (Constants.USE_WCET){
						transition.addUpdate("wcet = 0");
					}
				}
			} else if(!Options.EXPLICIT_MODEL && Options.PLATFORM_TYPE == PlatformType.HVMAVR) {
				if (Constants.USE_WCET){
					start.setInvariant("executionTime <= 0 &&\nexecutionTime' == running[tID] &&\nwcet' == running[tID]");
				} else {
					start.setInvariant("executionTime <= 0 &&\nexecutionTime' == running[tID]");
				}
				for (Transition transition : template.getOutgoingTransitions(start)) {
					//use set instead of add, because the old value must be overwritten.
					transition.setGuard("executionTime == 0");
					if (Constants.USE_WCET){
						transition.addUpdate("wcet = 0");
					}
				}
			}

			Transition endTransition = new Transition(end.getId(), initial.getId());
			endTransition.setSync(new Channel(channelName, methodInfo.getCanonicalName(), "tID"), true);
			
			if(!Options.EXPLICIT_MODEL) {
				if(Options.PLATFORM_TYPE == PlatformType.JOP) {
					String wcetInst = "inst" + end.getId();
					endTransition.addGuard("executionTime == " + (Constants.INLINE_WCET?end.getWCET():wcetInst));
				} else if(Options.PLATFORM_TYPE == PlatformType.HVMAVR) {
					endTransition.addGuard("executionTime == 0");
				}
				endTransition.addUpdate("executionTime = 0");
			}
			template.addTransition(initialTransition);
			template.addTransition(endTransition);
		}		
		uppaalModel.addTemplate(template);
	}
	
	private Hashtable<AbstractBasicBlock, Location> generateMethodBody(UppaalTemplate template, BasicBlockGraph basicBlockGraph){
		Hashtable<AbstractBasicBlock, Location> locationMap = new Hashtable<AbstractBasicBlock, Location>();
		
		LocationCreaterVisitor lcVisitor = new LocationCreaterVisitor(template, locationMap);
		basicBlockGraph.accept(lcVisitor);
		
		TransitionCreaterVisitor tcVisitor = new TransitionCreaterVisitor(template, locationMap, basicBlockGraph, classGraph);
		basicBlockGraph.accept(tcVisitor);
		
		TemplateDecoratorVisitor tdVisitor = new TemplateDecoratorVisitor(template, locationMap, basicBlockGraph, classGraph, sporadicTasks);
		basicBlockGraph.accept(tdVisitor);
		
		if(Options.CONSIDER_FINALIZERS || Options.COUNT_NEWS){
			TemplateNewCountVisitor newCountVisitor = new TemplateNewCountVisitor(template, locationMap, basicBlockGraph, classGraph, uppaalModel);
			basicBlockGraph.accept(newCountVisitor);
		}
		if(Options.CONSIDER_FINALIZERS){
			TemplateFinalizerCountVisitor finCounter = new TemplateFinalizerCountVisitor(template, locationMap, basicBlockGraph, classGraph, uppaalModel);
			basicBlockGraph.accept(finCounter);
		}
			
		
		tdVisitor.generateLoopBacktrack();
		
		return locationMap;
	}
}
