package tetasarts.uppaal.translation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import tetasarts.Options;
import tetasarts.generator.Constants;
import tetasarts.tir.IntermediateData;
import tetasarts.tir.classinfo.PeriodicTask;
import tetasarts.tir.classinfo.SporadicTask;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.translation.analysis.CHASootDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.IDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.NaiveDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.RTAAndVTASootDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.RTASootDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.VTASootDevirtualizationAnalysis;
import tetasarts.uppaal.translation.tasks.ConcreteMethodDecorator;
import tetasarts.uppaal.translation.tasks.TaskInfoFinder;

public class UppaalTranslation {
		
	public static UppaalModel generateUppaalModel(IntermediateData intermediateModel){
		UppaalModel model = new UppaalModel();
		ArrayList<ThreadDeadline> threadDeadlines = new ArrayList<ThreadDeadline>();
		TaskInfoFinder tif = new TaskInfoFinder(intermediateModel.getClassGraph(), intermediateModel.getMainClass());
		
		Set<PeriodicTask> periodicTasks = tif.getPeriodicTasks();
		Set<SporadicTask> sporadicTasks = tif.getSporadicTasks();
		
		IDevirtualizationAnalysis devirtualizer = null;
		
		switch(Options.DEVIRTUALIZATION_METHOD) {
			case NAIVE:
				devirtualizer = new NaiveDevirtualizationAnalysis(intermediateModel.getClassGraph());
				break;
			case CHA:
				devirtualizer = new CHASootDevirtualizationAnalysis(
						intermediateModel.getMainClass().getCanonicalName(),
						Options.MAIN_ENTRY_POINT, 
						intermediateModel.getClassGraph());
				break;
			case RTA:
				devirtualizer = new RTASootDevirtualizationAnalysis(
						intermediateModel.getMainClass().getCanonicalName(),
						Options.MAIN_ENTRY_POINT, 
						intermediateModel.getClassGraph());
				break;
			case VTA:
				devirtualizer = new VTASootDevirtualizationAnalysis(
						intermediateModel.getMainClass().getCanonicalName(),
						Options.MAIN_ENTRY_POINT, 
						intermediateModel.getClassGraph());
				break;
			case RTAVTA:
				devirtualizer = new RTAAndVTASootDevirtualizationAnalysis(
						intermediateModel.getMainClass().getCanonicalName(),
						Options.MAIN_ENTRY_POINT, 
						intermediateModel.getClassGraph());
				break;
			default:
				String msg = "Not capable of setting the devirtualization technique.";
				System.err.println(msg);
				System.exit(-1);
		}
		
		assert devirtualizer != null : "devirtualizer is null";
		
		ConcreteMethodDecorator cmd = new ConcreteMethodDecorator(devirtualizer);
		
		//Update periodicThread ID
		for (PeriodicTask periodicTask : periodicTasks) {
			model.nrPeriodicThreads++;
			periodicTask.setTID(model.nrPeriodicThreads);
			threadDeadlines.add(new ThreadDeadline(
					periodicTask.getDeadline(), 
					periodicTask.getTID()));

			cmd.decorateMethods(periodicTask.getInstanceOf().getMethod("run", Constants.RUN_METHOD_SIGNATURE), intermediateModel.getClassGraph());
		}
		//Update sporadicThread ID
		for (SporadicTask sporadicTask : sporadicTasks) {
			model.nrSporadicThreads++;
			sporadicTask.setTID(model.nrPeriodicThreads + model.nrSporadicThreads);
			threadDeadlines.add(new ThreadDeadline(
					sporadicTask.getDeadline(), 
					sporadicTask.getTID()));

			cmd.decorateMethods(sporadicTask.getInstanceOf().getMethod("run", Constants.RUN_METHOD_SIGNATURE), intermediateModel.getClassGraph());
		}
		model.setPeriodicTasks(periodicTasks);
		model.setSporadicTasks(sporadicTasks);
		
		if (!Constants.DISABLE_TASKS){	
			int priority = model.nrPeriodicThreads+model.nrSporadicThreads;
			//Test whether any periodic or sporadic threads exists
			if (priority > 0){
				Collections.sort(threadDeadlines);
				int priorities[] = new int[priority];
				//Shortest deadline gets the highest priority
				for (ThreadDeadline threadDeadline : threadDeadlines) {
					priorities[threadDeadline.tID-1] = priority;
					priority--;
				}
				StringBuilder sb = new StringBuilder();
				sb.append("int threadPriority[ThreadID] = {");
				sb.append(priorities[0]);
				for (int i = 1; i < priorities.length; i++){
					sb.append(", ").append(priorities[i]);
				}
				sb.append("};");
				model.getDeclarations().addString(sb.toString());
			} else {
				model.getDeclarations().addString("int threadPriority[ThreadID];");
			}
		}
		
		TemplateCreatorClassVisitor classVisitor = new TemplateCreatorClassVisitor(model, intermediateModel.getClassGraph(), sporadicTasks);
		intermediateModel.getClassGraph().accept(classVisitor);	
		
		return model;
	}
}

