package tetasarts.uppaal.translation.analysis;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.util.ClassPath;


import soot.PatchingChain;
import soot.Scene;
import soot.SootClass;
import soot.SootMethod;
import soot.Unit;
import soot.jimple.InvokeStmt;
import soot.jimple.VirtualInvokeExpr;
import soot.jimple.internal.JInvokeStmt;
import soot.jimple.toolkits.callgraph.CHATransformer;
import soot.jimple.toolkits.callgraph.CallGraph;
import soot.jimple.toolkits.callgraph.Edge;
import soot.jimple.toolkits.callgraph.Sources;
import soot.options.Options;
import soot.tagkit.BytecodeOffsetTag;
import soot.tagkit.LineNumberTag;
import soot.tagkit.Tag;
import soot.util.SingletonList;
import sun.awt.image.ImageWatched.Link;
import tetasarts.tir.basicblocks.basicBlockTypes.MethodCallingBasicBlock;
import tetasarts.tir.classinfo.AbstractClass;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.translation.analysis.bceltosoot.BcelToSootHelper;


public abstract class SootDevirtualizationAnalysis implements IDevirtualizationAnalysis {
	
	private final CallGraph callGraph;
	private ClassGraph classGraph;
	
	private static final String[] soot_class_path = ClassPath.getClassPath().split(":");
	
	private String getSootClasspath() {	
		StringBuilder sb = new StringBuilder();
		for(String path : SootDevirtualizationAnalysis.soot_class_path) {
			sb.append(path).append(':');
		}
		return sb.toString().substring(0, sb.length()-1);
	}
	
	abstract protected void setSootOptions();
	
	abstract protected void initDevirtualizationAnalysis();
	
	public SootDevirtualizationAnalysis(String mainClass, String mainMethod, ClassGraph classGraph) {
		this.setSootOptions();
		
		Scene.v().setSootClassPath(this.getSootClasspath());
		
		SootClass a = Scene.v().loadClassAndSupport(mainClass);
		SootMethod method = a.getMethodByName(mainMethod);
		
		List<SootMethod> entryPoints = new ArrayList<SootMethod>();
		entryPoints.add(method);
		Scene.v().setEntryPoints(entryPoints);
		Scene.v().loadNecessaryClasses();
		
		this.initDevirtualizationAnalysis();
		
		this.callGraph = Scene.v().getCallGraph();
		this.classGraph = classGraph;
	}
	
	private Unit getSootUnitAtByteCodeOffset(SootMethod sootMethod, int bytecodeOffset) {
		PatchingChain<Unit> chain = sootMethod.getActiveBody().getUnits();
		for(Unit stmt : chain) {
			/*
			 * This is extremely weird, and is a documented issue in the soot API. A jimple statement
			 * Can have MULTIPLE bytecodeoffsettags associated. Logically, it does not make sense, but we
			 * need to iterate over the possibilities.
			 */
			for(Tag tag : stmt.getTags()) {
				if(tag.getName().equals("BytecodeOffsetTag")) {
					int offset = ((BytecodeOffsetTag) tag).getBytecodeOffset();
					if(offset == bytecodeOffset)
						return stmt;
				}
			}
		}
		return null;
	}
	
	@Override
	public Set<MethodInformation> devirtualizeMethodCall(CallSite callSite) {
		Set<MethodInformation> candidateCallReceivers = new HashSet<MethodInformation>();
		
		/*
		 * TODO:
		 * this does not account for the possibility of overloaded methods!!! 
		 * Use the signature information in the callsite object to concretise this!
		 */
		InvokeInstruction callsiteInstruction = (InvokeInstruction)callSite.getInvokeInstructionHandle().getInstruction();
		ConstantPoolGen conPoolGen = new ConstantPoolGen(callSite.getCallSiteMethod().getConstantPool());
		
		String calleeMethodName = callsiteInstruction.getMethodName(conPoolGen);
		String calleeMethodSignature = callsiteInstruction.getSignature(conPoolGen);	
		String calleeClassName = ((ObjectType) callsiteInstruction.getReferenceType(conPoolGen)).getClassName();
		
		
		/*List<org.apache.bcel.generic.Type> bcelArgumentTypes = Arrays.asList(org.apache.bcel.generic.Type.getArgumentTypes(calleeMethodSignature));
		List<soot.Type> sootArgumentTypes = new LinkedList<soot.Type>();
		for(org.apache.bcel.generic.Type bcelT : bcelArgumentTypes) {
			sootArgumentTypes.add(BcelToSootHelper.bcelTypeToSootType(bcelT));
		}
		*/
		List<org.apache.bcel.generic.Type> bcelArgumentTypes = Arrays.asList(org.apache.bcel.generic.Type.getArgumentTypes(callSite.getCallSiteMethod().getSignature()));
		List<soot.Type> sootArgumentTypes = new LinkedList<soot.Type>();
		for(org.apache.bcel.generic.Type bcelT : bcelArgumentTypes) {
			sootArgumentTypes.add(BcelToSootHelper.bcelTypeToSootType(bcelT));
		}

		SootMethod sootMeth = Scene.v().getSootClass(callSite.getCallSiteClassName())
				.getMethod(callSite.getCallSiteMethod().getName(), sootArgumentTypes);
		
		int positionOfCallSiteInvoke = returnPositionOfInvokeStmt(callSite.getInvokeInstructionHandle());
	
		Unit sootUnit = getSootUnitAtByteCodeOffset(sootMeth, positionOfCallSiteInvoke);
		
	//	if(sootUnit instanceof InvokeStmt) {
			Iterator<Edge> outgoingEdgesIter = this.callGraph.edgesOutOf(sootUnit);
			
			while(outgoingEdgesIter.hasNext()) {
				Edge edge = outgoingEdgesIter.next();
				SootMethod sootTargetMethod = edge.getTgt().method();
				MethodInformation sirTargetMethod = null;
				try {
					sirTargetMethod = getSirMethodInfoFromSootMethod(sootTargetMethod);
				} catch (Exception e) {
					e.printStackTrace();
					System.exit(1);
				}
				
				assert sirTargetMethod != null : "could not translate from soot to sir";
				
				candidateCallReceivers.add(sirTargetMethod);
				
				if(callSite.isStaticInvoke())
					return candidateCallReceivers;
			}
		//}
		return candidateCallReceivers;
	}
	
	
	private MethodInformation getSirMethodInfoFromSootMethod(SootMethod method) throws Exception {
		AbstractClassNode classNode = this.classGraph.getClassFromCanonicalName(method.getDeclaringClass().getName());
		
		String fullyQualifiedMethodSignature = method.getBytecodeSignature();
		String byteCodeMethodSignature = extractByteCodeSignature(fullyQualifiedMethodSignature);
		MethodInformation sirMethod = classNode.getMethod(byteCodeMethodSignature);
		if(sirMethod == null) 
			throw new Exception("Could not translate soot method [" + fullyQualifiedMethodSignature + "] to sir method");
		return sirMethod;
	}
	
	private String extractByteCodeSignature(String sootBytecodeSignature) {
		return sootBytecodeSignature.substring(sootBytecodeSignature.indexOf(':') + 1, sootBytecodeSignature.length() - 1).trim();
	}
	
	private int returnPositionOfInvokeStmt(InstructionHandle invokeInstruction) {
		if(invokeInstruction.getInstruction() instanceof InvokeInstruction) {
			return invokeInstruction.getPosition();
		}else {
			return -1;
		}
	}
	
	private List<Integer> returnPositionsOfInvokeStmt(Unit sootInstruction) {
		JInvokeStmt invokeStmt = (JInvokeStmt)sootInstruction;
		
		List<Integer> bytecodeOffsets = new LinkedList<Integer>();
		for(Tag tag : invokeStmt.getTags()) {
			if(tag.getName().equals("BytecodeOffsetTag")) {
				bytecodeOffsets.add(((BytecodeOffsetTag) tag).getBytecodeOffset());
			}
		}
		
		return bytecodeOffsets;
		
		
		/*
		int offset = Integer.MAX_VALUE;
		int newOffset = Integer.MAX_VALUE;
		for(Tag offsetTag : bytecodeOffsetTags) {
			if(offsetTag.getName().equals("BytecodeOffsetTag")) {
				newOffset = ((BytecodeOffsetTag) offsetTag).getBytecodeOffset();
				if(newOffset < offset) {
					offset = newOffset;
				}
			}
		}
		*/
	//	return newOffset;
	}
	
	
}
