package com.tetaj.cfggenerators;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.ControlFlowGraphException;
import com.tetaj.ircfg.Instruction;

import cck.util.Option;
import cck.util.Options;

import avrora.Defaults;
import avrora.arch.legacy.LegacyInstr;
import avrora.core.ControlFlowGraph.Edge;
import avrora.core.ProcedureMap;
import avrora.core.Program;
import avrora.core.ProgramReader;
import avrora.core.SourceMapping;
import avrora.core.ControlFlowGraph.Block;
import avrora.core.SourceMapping.Location;

public class AVRToIRCFG implements IIRCFGGenerators {
	
	private class Addr2LineResult {
		public String procedureName;
		public String absolutePath;
		public int lineNumber;
	}
	
	private class Addr2Line {
	    private Process addr2line;
	    private BufferedWriter stdin;
	    private BufferedReader stdout;
	    private BufferedReader stderr;
	    
	    public Addr2Line(String executable) throws Addr2lineException {
			String[] args = { "avr-addr2line", "-C", "-f", "-e", executable};
			try {
				addr2line = Runtime.getRuntime().exec(args);
			} catch (IOException e) {
				throw new Addr2lineException("Could not execute avr-addr2line properly.");
			}
			stdin = new BufferedWriter(new OutputStreamWriter(addr2line.getOutputStream()));
			stdout = new BufferedReader(new InputStreamReader(addr2line.getInputStream()));
			stderr = new BufferedReader(new InputStreamReader(addr2line.getErrorStream()));
			
	    }
	    
	    public Addr2LineResult obtainByteOffsetInfoDecimal(long byteOffset) throws Addr2lineException {
	    	String source = "";
	    	Addr2LineResult result = new Addr2LineResult();
	    	try {
				stdin.write(Long.toHexString(byteOffset) + "\n");
				stdin.flush();
		    	result.procedureName = stdout.readLine();
				source = stdout.readLine();
			} catch (IOException e) {
				throw new Addr2lineException("avr-addr2line could not obtain information about the instruction with offset: " + byteOffset);
			} 
			String[] pathAndNum = source.split(":");
			result.absolutePath = pathAndNum[0];
			result.lineNumber = Integer.parseInt(pathAndNum[1]);
			
			return result;
	    }
	}
	
	private static final Options mainOptions = new Options();
	private static final Option.Str INPUT = mainOptions.newOption("input", "auto",
            "This option selects among the available program formats as input to Avrora. " +
            "For example, the default input format, \"atmel\" selects the assembly " +
            "language format supported by Atmel's assembler.");
	
	private final String programLocation;
	private final Program program;
	private final Addr2Line addr2line;
	
	private static final String CALL_EDGE = "CALL";
	private static final String NORMAL_EDGE = "";
	
	private HashMap<Block, ControlFlowGraph> processedCFGs = new HashMap<Block, ControlFlowGraph>();
	private HashMap<Block, BasicBlock> processedBlocks = new HashMap<Block, BasicBlock>();
	private String fqMehtodname;
	
	public AVRToIRCFG(String fqMethodName, final String programLocation) throws Exception {
		ProgramReader reader = Defaults.getProgramReader(INPUT.get());
		reader.options.process(mainOptions);
		this.programLocation = programLocation;
		this.addr2line = new Addr2Line(programLocation);
        this.program = reader.read(new String[]{programLocation});
        this.fqMehtodname = fqMethodName;
	}

	@Override
	public HashMap<String, ControlFlowGraph> generateIRCFG() throws Addr2lineException, AVRToIRCFGException, ControlFlowGraphException {
		//Block mainStartBasicBlock = getMainStartBlock(this.program);
		
		SourceMapping srcMap = this.program.getSourceMapping();
		Location methodLoc = srcMap.getLocation(this.fqMehtodname);
		
		avrora.core.ControlFlowGraph cfg = this.program.getCFG();
		
		Block methodBlockEntry = cfg.getBlockStartingAt(methodLoc.lma_addr);
		buildCFGRecursively(methodBlockEntry);
		
		HashMap<String, ControlFlowGraph> procNameCFGMap = getCleanedCFGMap(processedCFGs);
		HashMap<String, ControlFlowGraph> strippedProcNameCFGMap = stripCfgHashMap(this.fqMehtodname, procNameCFGMap);
		
		return strippedProcNameCFGMap;
	}
	
	private HashMap<String, ControlFlowGraph> stripCfgHashMap(String fqMehtodname, HashMap<String, ControlFlowGraph> procNameCFGMap) {
		HashMap<String, ControlFlowGraph> strippedCfgs = new HashMap<String, ControlFlowGraph>();
		
		ControlFlowGraph initialCfg = procNameCFGMap.get(fqMehtodname);
		
		recursivelyAddMethodInvokeCfg(initialCfg, strippedCfgs);
		
		
		return strippedCfgs;
	}

	private void recursivelyAddMethodInvokeCfg(ControlFlowGraph cfg, HashMap<String, ControlFlowGraph> strippedCfgs) {
		if (!strippedCfgs.containsKey(cfg.getMethodName())) {
			strippedCfgs.put(cfg.getMethodName(), cfg);
			
			for (BasicBlock invokeBb : cfg.getMethodInvokeList()) {
				for (ControlFlowGraph invokeCfg : invokeBb.getMethodInvokeCfg().values()) {
					recursivelyAddMethodInvokeCfg(invokeCfg, strippedCfgs);
				}
			}
		}
	}

	private HashMap<String, ControlFlowGraph> getCleanedCFGMap(HashMap<Block, ControlFlowGraph> CFGs) {
		
		HashMap<String, ControlFlowGraph> cleanedCFGMap = new HashMap<String, ControlFlowGraph>(); 
		
		Collection<ControlFlowGraph> cleanedCFGs = cleanCFGforNulls(CFGs.values());

		// Fix possible name clashes
		for(ControlFlowGraph currentCFG : cleanedCFGs) {
			String hashString = currentCFG.getMethodName();
			int appendedId = 0;
			while(cleanedCFGMap.containsKey(hashString)) {
				hashString = currentCFG.getMethodName() + appendedId++;
			}
			currentCFG.setMethodName(hashString);
			cleanedCFGMap.put(hashString, currentCFG);
		}
		
		return cleanedCFGMap;
	}
	
	private HashSet<ControlFlowGraph> cleanCFGforNulls(Collection<ControlFlowGraph> CFGs) {
		HashSet<ControlFlowGraph> cleanedCfgs = new HashSet<ControlFlowGraph>();
		cleanedCfgs.addAll(CFGs);
		
		for (ControlFlowGraph cfg : CFGs) {	
			for (BasicBlock basicBlock : cfg.getBasicBlocks()) {
				
				if (basicBlock.isMethodInvoke()) {
					
					HashSet<ControlFlowGraph> invokedCfgs = new HashSet<ControlFlowGraph>();
					invokedCfgs.addAll(basicBlock.getMethodInvokeCfg().values());
					
					for (ControlFlowGraph invokeCfg : invokedCfgs) {
						if (invokeCfg.getBasicBlocks().size() == 0) {
							System.out.println("removing self method invoke " + invokeCfg.getMethodName());
							cleanedCfgs.remove(invokeCfg);
							basicBlock.getMethodInvokeCfg().remove(invokeCfg);
							basicBlock.getInstructions().get(basicBlock.getInstructions().size()-1).setMethodinvoke(false);
						}
					}	
				}
			}
		}
		
		return cleanedCfgs;
	}
	
	private ControlFlowGraph buildCFGRecursively(Block procEntryBlock) throws Addr2lineException, AVRToIRCFGException, ControlFlowGraphException {
		if(processedCFGs.containsKey(procEntryBlock)) {
			return processedCFGs.get(procEntryBlock);
		}

		Addr2LineResult addr2lineResult = this.addr2line.obtainByteOffsetInfoDecimal(procEntryBlock.getAddress());
		String sourcePath = addr2lineResult.absolutePath;
		
		if(sourcePath.contains("?"))
			throw new AVRToIRCFGException("Executable was not compiled with debugging information. Probably missed the -g option");
    	
		String procName = addr2lineResult.procedureName;
		
		if(procName.contains("?"))
			throw new AVRToIRCFGException("Executable was not compiled with debugging information. Probably missed the -g option");
		
		System.out.println("added proc: " + procName);
		ControlFlowGraph tetajCFG = new ControlFlowGraph(sourcePath, procName);
		translateAvroraBlockToTetajRecursively(tetajCFG, procEntryBlock);
		
		processedCFGs.put(procEntryBlock, tetajCFG);
		
		return tetajCFG;
	}
	
	private BasicBlock translateAvroraBlockToTetajRecursively(ControlFlowGraph tetajCFG, Block procEntryBlock) throws Addr2lineException, AVRToIRCFGException, ControlFlowGraphException {
		if(processedBlocks.containsKey(procEntryBlock)) {
			return processedBlocks.get(procEntryBlock);
		}
		BasicBlock tetajSourceBasicBlock = translateAvroraBlockToTetaj(procEntryBlock, tetajCFG.highestBlockIndex);
		tetajCFG.highestBlockIndex++;
		tetajCFG.addBasicBlock(tetajSourceBasicBlock);
		processedBlocks.put(procEntryBlock, tetajSourceBasicBlock);
		Iterator<Edge> avroraEdgeIterator = (Iterator<Edge>)procEntryBlock.getEdgeIterator();
		while(avroraEdgeIterator.hasNext()) {
			Edge avroraEdge = avroraEdgeIterator.next();
			if(avroraEdge.getTarget() != null) {	
				if(avroraEdge.getType().equals(NORMAL_EDGE)) {
					Block avroraSourceBlock = avroraEdge.getSource();
					Block avroraTargetBlock = avroraEdge.getTarget();
					BasicBlock tetajTargetBasicBlock = translateAvroraBlockToTetajRecursively(tetajCFG, avroraTargetBlock);
					/*
					 * TODO: remove the notion of branch and fallthrough basic blocks. The distinction is not
					 * used anymo'
					 */
					tetajSourceBasicBlock.addBranch(tetajTargetBasicBlock);
					
					/*if(isTargetAvroraBlockBranch(avroraSourceBlock, avroraTargetBlock)) {	
						tetajSourceBasicBlock.addBranch(tetajTargetBasicBlock);
					}
					else {
						tetajSourceBasicBlock.setFallthrough(tetajTargetBasicBlock);
					}*/
	
				} else if(avroraEdge.getType().equals(CALL_EDGE)) {
					ArrayList<Instruction> instructions = tetajSourceBasicBlock.getInstructions();
					instructions.get(instructions.size()-1).setMethodinvoke(true);
						ControlFlowGraph c = buildCFGRecursively(avroraEdge.getTarget());
						tetajSourceBasicBlock.addMethodInvokeCfg(c, tetajSourceBasicBlock.getInstructions().size()-1); //Only one method invoke for AVR
				}
			}
		}
		return tetajSourceBasicBlock;
	}
	
	private Block getMainStartBlock(Program program) throws Addr2lineException, AVRToIRCFGException {
		SourceMapping sourceMap = program.getSourceMapping();
		Iterator<Location> sourceMapIter = (Iterator<Location>)sourceMap.getIterator();
		int mainEntryPointByteLocation = 0;
		while(sourceMapIter.hasNext()) {
			Location procEntryPoint = sourceMapIter.next();
			String procName = this.addr2line.obtainByteOffsetInfoDecimal(procEntryPoint.lma_addr).procedureName;
			if(procName.equals("main")) {
				mainEntryPointByteLocation = procEntryPoint.lma_addr;
				break;
			}
		}
		if(mainEntryPointByteLocation == 0) {
			throw new AVRToIRCFGException("Could not locate the main function");
		}
		else {
			avrora.core.ControlFlowGraph programCFG = program.getCFG();
			Iterator<Block> programBasicBlocksIter = programCFG.getSortedBlockIterator();
			while(programBasicBlocksIter.hasNext()) {
				Block currentBlock = programBasicBlocksIter.next();
				if(currentBlock.getAddress() == mainEntryPointByteLocation) {
					return currentBlock;
				}
			}
		}
		return null;
	}
	
	private boolean isTargetAvroraBlockBranch(Block avroraSourceBlock, Block avroraTargetBlock) {
		long sourceStartAddr = avroraSourceBlock.getAddress();
		long sourceLastAddr = avroraSourceBlock.getLastAddress();
		long targetStartAddr = avroraTargetBlock.getAddress();
		long targetLastAddr = avroraTargetBlock.getLastAddress();
		if(targetLastAddr < sourceLastAddr)
			return true;

		/*
		 * This is the greatest hack in the world
		 */
		LegacyInstr firstInstrOfTarget = null;
		if(avroraTargetBlock.getInstrIterator().hasNext()) {
			firstInstrOfTarget = (LegacyInstr)avroraTargetBlock.getInstrIterator().next();
			if(sourceLastAddr + firstInstrOfTarget.getSize() < targetStartAddr) {
				return true;
			}
			else {
				return false;
			}
		}
		/*
		 * Not sure whether following "else" is correct...
		 */
		else {
			return false;
		}
		
	}
	
	private BasicBlock translateAvroraBlockToTetaj(Block avroraBasicBlock, int basicBlockIndex) throws Addr2lineException {
		BasicBlock tetajBasicBlock = new BasicBlock(basicBlockIndex);
		long byteOffset = avroraBasicBlock.getAddress();
		Iterator<LegacyInstr> avroraInstructionIterator = (Iterator<LegacyInstr>)avroraBasicBlock.getInstrIterator();
		while(avroraInstructionIterator.hasNext()) {
			LegacyInstr avroraInstruction = avroraInstructionIterator.next();
			Instruction tetajInstruction = translateAvroraInstruction(avroraInstruction, byteOffset);
			tetajBasicBlock.addInstruction(tetajInstruction);
			/*
			 * check whether the basicblock address equals the byteoffset of the first
			 * instruction. If it is, following might be wrong
			 */
			byteOffset += avroraInstruction.getSize();
		}
		
		return tetajBasicBlock;
	}
	
	private Instruction translateAvroraInstruction(LegacyInstr avroraInstruction, long avroraByteOffset) throws Addr2lineException {
		/*
		 * Should be the opcode and not the mnemonic that is used in the constructor...
		 */
		Instruction tetajInstruction = new Instruction(avroraInstruction.getName());
		Addr2LineResult addr2lineResult = this.addr2line.obtainByteOffsetInfoDecimal(avroraByteOffset);	
		tetajInstruction.setLinenumber(addr2lineResult.lineNumber);
		tetajInstruction.setMnemonic(avroraInstruction.getName());
		
		String tetajInstructionOperands = "";
		
		if(isInstructionCall(avroraInstruction)) {
			String procedureName = getCalledProcedureName(avroraInstruction, avroraByteOffset);
			tetajInstructionOperands = avroraInstruction.getOperands() + " (" + procedureName + ")";
		}
		else
			tetajInstructionOperands = avroraInstruction.getOperands();
		
		tetajInstruction.setOperands(tetajInstructionOperands);
		

		return tetajInstruction;
	}
	
	private String getCalledProcedureName(LegacyInstr avroraInstruction, long byteLocation) throws Addr2lineException {
		/*
		 * please note the multiplication with 2. For some very odd reason,
		 * avrora specifies the half(!!) byteoffset in the operand of rcall instructions.
		 */
		String procedureName;
		if(!avroraInstruction.getOperands().equals("")) {
			long byteOffsetOfProcedure = Integer.parseInt(avroraInstruction.getOperands()) * 2 + avroraInstruction.getSize();
			long absoluteLocationOfProcedure = byteLocation + byteOffsetOfProcedure;
			Addr2LineResult addr2lineResult = this.addr2line.obtainByteOffsetInfoDecimal(absoluteLocationOfProcedure);
			procedureName = addr2lineResult.procedureName;
		}
		else {
			procedureName = "NOT DETERMINED";
		}

		return procedureName;
	}
	
	private boolean isInstructionCall(LegacyInstr avroraInstruction) {
		String avroraInstructionType = avroraInstruction.getVariant();
		if(avroraInstructionType.equals("rcall") || 
		   avroraInstructionType.equals("call") ||
		   /*
		    * eicall and icall should probably not be accounted for..
		    */
		   avroraInstructionType.equals("icall") ||
		   avroraInstructionType.equals("eicall")) {
			return true;
		}
		else
			return false;
	}
	
}
