package com.tetaj.hvmhelper;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.ResourceBundle.Control;

import com.tetaj.cfggenerators.AVRToIRCFG;
import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.ControlFlowGraphException;
import com.tetaj.ircfg.Edge;
import com.tetaj.ircfg.Instruction;
import com.tetaj.ircfg.LoopInfo;

public class HvmCfgHelper {

	public static HashMap<String, ControlFlowGraph> generateCfgsForHvm(String fqMethodName, String elfPath, String[] instructions, String methodPrefix) throws Exception{
		HashMap<String, ControlFlowGraph> cfgs = new HashMap<String, ControlFlowGraph>();
		
		
		AVRToIRCFG avrGenerator = new AVRToIRCFG(fqMethodName, elfPath);
		cfgs = avrGenerator.generateIRCFG();
				
		for(String instruction : instructions){
			if(instruction!=Constants.ILLEGAL_OPCODE){
				avrGenerator = new AVRToIRCFG(getMethodName(instruction,methodPrefix), elfPath);
				HashMap<String, ControlFlowGraph> instructionCfgs = avrGenerator.generateIRCFG();
				
				mergeHashMaps(cfgs,instructionCfgs);
			}
		}
		
		return cfgs;
	}

	public static void handleInterpretationLoopCfg(HashMap<String, ControlFlowGraph> cfgs) throws Exception {

		ControlFlowGraph combinedCfg = cfgs.get("process_jbc");
		
		BasicBlock cuttinBasicBlock = null;
		LoopInfo whileLoop = null;
		
		loop:
		for (BasicBlock basicBlock : combinedCfg.getBasicBlocks()) {
			
			if (basicBlock.hasProperty("loopinfo")) {
				LoopInfo loopInfo = (LoopInfo)basicBlock.getProperty("loopinfo");
				
				for (BasicBlock loopBasicBlock: loopInfo.getBasicBlocks()) {
					for (Instruction instr : loopBasicBlock.getInstructions()) {
						
						if (instr.getMnemonic() == "eicall") {
							whileLoop = loopInfo;
							cuttinBasicBlock = loopBasicBlock;
							
							if (loopBasicBlock.getInstructions().indexOf(instr) != 
								loopBasicBlock.getInstructions().size() - 1) {
								// eicall was not the last instruction in block
								throw new Exception("EICall was not the last instruction in the basic block as expected");
							}
						
							break loop;
						}
					}
				}
			}
		}
				
		if (cuttinBasicBlock == null){
			throw new Exception("No Eicall was found in HVM");
		}
		
		ControlFlowGraph preCfg = new ControlFlowGraph(combinedCfg.getSourceFilePath(), combinedCfg.getMethodName() + "__pre__");
		ControlFlowGraph postCfg = new ControlFlowGraph(combinedCfg.getSourceFilePath(), combinedCfg.getMethodName() + "__post__");
				
		splitCFG(combinedCfg, preCfg, postCfg, cuttinBasicBlock, whileLoop);
		
		cfgs.put(preCfg.getMethodName(), preCfg);
		cfgs.put(postCfg.getMethodName(), postCfg);
		
		fixRecursiveMethodInterpreterCall(cfgs);
		cfgs.remove("process_jbc");
		
	}

	private static void fixRecursiveMethodInterpreterCall(HashMap<String, ControlFlowGraph> cfgs) {
			
		replaceCfgWithEmptyCfg("createStringObject",cfgs);
		
		
	}
	
	private static void replaceCfgWithEmptyCfg(String methodName, HashMap<String, ControlFlowGraph> cfgs){
		try {
		String sourcePath = cfgs.get(methodName).getSourceFilePath();
		cfgs.remove(methodName);
		ControlFlowGraph cfgCreateString = new ControlFlowGraph(sourcePath, methodName);
		cfgCreateString.addBasicBlock(new BasicBlock(0));
		cfgs.put(methodName, cfgCreateString);	
		}
		catch (NullPointerException e) {
			
		}
	}

	@SuppressWarnings("unchecked")
	private static void splitCFG(ControlFlowGraph combinedCfg, ControlFlowGraph preCfg, ControlFlowGraph postCfg,
			BasicBlock cuttingBasicBlock, LoopInfo whileLoop) throws Exception {
		
		// move all basic blocks before cuttingBasicBlock into pre CFG
		BasicBlock loopHeader = whileLoop.getBackEdges().get(0).getTarget();
		loopHeader.setId(0);
		
		LinkedList<BasicBlock> queue = new LinkedList<BasicBlock>();
		queue.push(loopHeader);
		
		while (queue.size() != 0) {
			BasicBlock current = queue.pop();
			preCfg.addBasicBlock(current);
			
			if (current == cuttingBasicBlock) {
				continue;
			}
			
			for (Edge edge : current.getOutgoingEdges()) {
				queue.push(edge.getTarget());
			}
		}
		
		// move all basic blocks after cuttingBasicBlock into post CFG
		BasicBlock entryBasicBlock = cuttingBasicBlock.getOutgoingEdges().get(0).getTarget();
		entryBasicBlock.setId(0);
		
		queue.clear();
		queue.push(entryBasicBlock);
		
		if (cuttingBasicBlock.getOutgoingEdges().size() > 1) {
			throw new Exception("Branches on eicall block not expected");
		}
		
		while (queue.size() != 0) {
			BasicBlock current = queue.pop();
			postCfg.addBasicBlock(current);

			for (Edge edge : current.getOutgoingEdges()) {
				if (!edge.isLoopIteration() && !edge.isLoopExit() && 
						!postCfg.getBasicBlocks().contains(edge.getTarget()) &&
						edge.getTarget() != cuttingBasicBlock) {
					
					queue.push(edge.getTarget());
				}
			}
		}
		
		cleanUnusedEdges(preCfg);
		cleanUnusedEdges(postCfg);
		
		preCfg.resetIds();
		postCfg.resetIds();
	}
	
	@SuppressWarnings("unchecked")
	private static void cleanUnusedEdges(ControlFlowGraph cfg) {
		for (BasicBlock basicBlock : cfg.getBasicBlocks()) {
			
			for (Edge edge : (ArrayList<Edge>)basicBlock.getOutgoingEdges().clone()) {
				if (!cfg.getBasicBlocks().contains(edge.getTarget())) {
					basicBlock.removeEdge(edge);
				}
			}
			
			for (Edge edge : (ArrayList<Edge>)basicBlock.getIncomingEdges().clone()) {
				if (!cfg.getBasicBlocks().contains(edge.getSource())) {
					edge.getSource().removeEdge(edge);
				}
			}
		}
	}

	private static void mergeHashMaps(HashMap<String, ControlFlowGraph> mergeInto, HashMap<String, ControlFlowGraph> mergeFrom) {
		for(String key : mergeFrom.keySet()){
			if(!mergeInto.containsKey(key))
				mergeInto.put(key, mergeFrom.get(key));		
		}		
	}

	private static String getMethodName(String instruction,String methodPrefix) {
		String methodName = methodPrefix + instruction.toUpperCase()+"_OPCODE";
		
		return methodName;
	}

	public static void generateNativeCfgs(String elfPath, String[] nativeMethods,	HashMap<String, ControlFlowGraph> cfgtable) throws Exception {
		
		AVRToIRCFG avrGenerator;
		HashMap<String, ControlFlowGraph> newCfgs;
		for(String nativeMethod : nativeMethods){
			try{
				avrGenerator = new AVRToIRCFG(nativeMethod, elfPath);
				newCfgs = avrGenerator.generateIRCFG();
				mergeHashMaps(cfgtable, newCfgs);
				System.out.println("Found Native Method: " + nativeMethod);
			}catch(Exception e){
				e.printStackTrace();
				throw new Exception("Could not add native method " + nativeMethod + " to the list of CFGs");
			}
		}
	}
	
}
