package com.tetaj.modelgenerator.cfganalysis;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.ControlFlowGraphException;
import com.tetaj.ircfg.Edge;
import com.tetaj.ircfg.LoopInfo;

import com.tetaj.modelgenerator.annotationextractor.extractor.CStyleAnnotationExtractor;
import com.tetaj.modelgenerator.annotationextractor.parser.Annotation;
import com.tetaj.modelgenerator.annotationextractor.parser.Annotation.AnnotationType;

public class LoopBoundAnalysis implements IAnalysis {
	
	private static final String VISIT_PROPERTY = "visit";
	private static final String DFST_BASIC_BLOCK_ID_PROPERTY = "dfstvalue";
	private static final String VISITED = "v";
	private static final String UNVISITED = "uv";
	private static final String BACKEDGE = "backedge";
	private static final int CFG_ENTRY_BASIC_BLOCK = 0;
	
	private static final String LOOP_INFO_PROPERTY = "loopinfo";
	
	private int globalBasicBlockCount;
	
	private LinkedList<Edge> backEdges;
	private HashMap<String, ControlFlowGraph> CFGtable;
	
	public LoopBoundAnalysis(HashMap<String, com.tetaj.ircfg.ControlFlowGraph> CFGTable) {
		this.CFGtable = CFGTable;
	}
	
	@Override
	public void conductAnalysis() throws CfgAnalysisException {
		ArrayList<ControlFlowGraph> CFGs = new ArrayList<ControlFlowGraph>();
		CFGs.addAll(CFGtable.values());
		
		//Clean empty Cfgs
		for(ControlFlowGraph cfg : CFGtable.values()) 
			if(cfg.getBasicBlocks().size()==0){
				CFGs.remove(cfg);
				System.out.println("Removing: " + cfg.getMethodName());
			}
			
		
		for(ControlFlowGraph cfg : CFGs) 
				conductDepthFirstOrdering(cfg);
		
		for (ControlFlowGraph CFG : CFGs) {
	
				String sourcePath = CFG.getSourceFilePath();
				
				LinkedList<Edge> backEdges = determineBackEdges(CFG);
				LinkedList<LoopInfo> foundLoops = new LinkedList<LoopInfo>(); 
				
				for (Edge backEdge : backEdges) {
					
					prepareCFGsForDFSAnalysis(CFG);
					constructLoopSet(backEdge, foundLoops);
				}
				
				for (LoopInfo loopInfo : foundLoops) {
					
					try {
						setEntryAndExitEdges(loopInfo, sourcePath);
					} catch (ControlFlowGraphException e) {
						e.printStackTrace();
						throw new CfgAnalysisException("Error in selecting entry and exit edges for loop");
					}
				}
	
			
		}
	}
	
	private int getLoopBound(BasicBlock basicBlock, String sourcePath) throws CfgAnalysisException {
		CStyleAnnotationExtractor annoExtractor = new CStyleAnnotationExtractor();
		
		int lineNumber = getLineNumberFromBasicBlock(basicBlock);
		
		try {
			LinkedList<Annotation> annotations = annoExtractor.retrieveAnnotations(sourcePath, lineNumber);
			
			if (annotations.size() == 0) {
				throw new CfgAnalysisException("No loopbounds set in file: " + sourcePath + " on linenumber: " + lineNumber);
			}
			
			for (Annotation annotation : annotations) {
				if (annotation.getAnnotationType() == AnnotationType.LOOPBOUND) {
					return Integer.parseInt(annotation.getAnnotationValue());
				}
			}
			
		} catch (IOException e) {
			e.printStackTrace();
			throw new CfgAnalysisException("Could not extract annotations");
		}
		
		if(true) {
			throw new CfgAnalysisException("No loopbounds set in file: " + sourcePath + " on linenumber: " + lineNumber);
		}
		return -1;
	}
	
	private int getLineNumberFromBasicBlock(BasicBlock basicBlock) {
		return basicBlock.getInstructions().get(0).getLinenumber();
	}
	
	private void setEntryAndExitEdges(LoopInfo loopInfo, String sourcePath) throws CfgAnalysisException, ControlFlowGraphException {
		
		LinkedList<Edge> exitEdges = getExitEdges(loopInfo);
		boolean foundCondition = false;
		
		for(Edge edge : exitEdges)
			edge.setAsLoopExit(loopInfo);
		
		BasicBlock loopHead = loopInfo.getBackEdges().get(0).getTarget();
		
		for (Edge edge : loopHead.getIncomingEdges()) {
			if (loopInfo.getBasicBlocks().contains(edge.getSource()) == false) {
				edge.setAsLoopEntry(loopInfo);
			}
		}
		
		for (BasicBlock basicBlock : loopInfo.getBasicBlocks()) {
			
			Edge pointsToLoop = null;
			Edge pointsOutOfLoop = null;
			
			for (Edge edge : basicBlock.getOutgoingEdges()) {
				if (exitEdges.contains(edge)) {
					pointsOutOfLoop = edge;
				} else {
					pointsToLoop = edge;
				}
			}
			
			if (pointsToLoop != null && pointsOutOfLoop != null) {
				foundCondition = true;
				
				try {
					int loopBound = this.getLoopBound(basicBlock, sourcePath);
					int linenum = this.getLineNumberFromBasicBlock(loopInfo.getBasicBlocks().get(0));
					
					loopInfo.setLoopBound(loopBound);
					pointsToLoop.setAsLoopIteration(loopInfo);
					
					//System.out.println("loop bound to be set: " + loopBound + " for loop in line " + linenum + " in file " + sourcePath);
					return;
					
				} catch (CfgAnalysisException e) {
					e.printStackTrace();
				}
			}
		}
		
		int linenumber = this.getLineNumberFromBasicBlock(loopInfo.getBasicBlocks().get(0));
		
		// support loops with no condition loop but a -1 loop annotation
		if (!foundCondition) {
			for (BasicBlock basicBlock : loopInfo.getBasicBlocks()) {
				try {
					int loopBound = this.getLoopBound(basicBlock, sourcePath);
					
					if (loopBound == -1) {
						System.out.println("WARNING: unbounded loop with no condtion block detected at line " + linenumber + " in file: " + sourcePath);
						return;
					}
				} catch (CfgAnalysisException e) {
					// ignore
				}
			}
		}
		
		throw new CfgAnalysisException("No annotation found for loop: line " + linenumber + " in file: " + sourcePath);
	}
	
	private LinkedList<Edge> getExitEdges(LoopInfo loopHeaderInfo) {
		LinkedList<Edge> exitEdges = new LinkedList<Edge>();
		
		ArrayList<BasicBlock> loopBasicBlocks = loopHeaderInfo.getBasicBlocks();
		
		for(BasicBlock basicBlock : loopBasicBlocks) {
			for(Edge edge : basicBlock.getOutgoingEdges()) {
				if(loopBasicBlocks.contains(edge.getTarget()) == false)
					exitEdges.add(edge);
			}
		}
		
		return exitEdges;
	}
	
	private void constructLoopSet(Edge backEdge, LinkedList<LoopInfo> foundLoops) {
		BasicBlock loopHeader = backEdge.getTarget();
		LoopInfo loopInfo = null;
		
		backEdge.getTarget().setProperty(VISIT_PROPERTY, VISITED);
		
		if (loopHeader.hasProperty(LOOP_INFO_PROPERTY)) {
			loopInfo = (LoopInfo) loopHeader.getProperty(LOOP_INFO_PROPERTY);
			
			if (loopInfo.getBackEdges().contains(backEdge) == false) {
				loopInfo.addBackEdge(backEdge);
			}
		
		} else {
			
			loopInfo = new LoopInfo();
			loopInfo.addBackEdge(backEdge);
			
			foundLoops.add(loopInfo);
			loopHeader.setProperty(LOOP_INFO_PROPERTY, loopInfo);
	
			loopInfo.addBasicBlock(backEdge.getTarget());
		}
		
		conductReverseDFS(backEdge.getSource(), loopInfo);
	}
	
	private void conductReverseDFS(BasicBlock currentBasicBlock, LoopInfo loopInfo) {
		
		if (currentBasicBlock.getProperty(VISIT_PROPERTY) == UNVISITED) {
			
			currentBasicBlock.setProperty(VISIT_PROPERTY, VISITED);
			loopInfo.addBasicBlock(currentBasicBlock);
			for(Edge incomingEdge : currentBasicBlock.getIncomingEdges()) {
					conductReverseDFS(incomingEdge.getSource(), loopInfo);
			}
			
		}
	}
	
	private LinkedList<Edge> determineBackEdges(ControlFlowGraph CFG) {
		this.backEdges = new LinkedList<Edge>();
		
		prepareCFGsForDFSAnalysis(CFG);
		BasicBlock CFGEntryBasicBlock = CFG.getBasicBlock(CFG_ENTRY_BASIC_BLOCK);
		searchForBackEdge(CFGEntryBasicBlock);
		return this.backEdges;
	}
	
	private void searchForBackEdge(BasicBlock currentBasicBlock) {
		currentBasicBlock.setProperty(VISIT_PROPERTY, VISITED);
		List<Edge> successorsOfCurrentBasicBlock = currentBasicBlock.getOutgoingEdges();
		for(Edge successorEdge : successorsOfCurrentBasicBlock) {
			BasicBlock targetBasicBlock = successorEdge.getTarget();
			
			int sourceBasicBlockDFSTID = getDFSTBasicBlockId(currentBasicBlock);
			int targetBasicBlockDFSTID = getDFSTBasicBlockId(targetBasicBlock);
			if(sourceBasicBlockDFSTID >= targetBasicBlockDFSTID) {
				if(this.backEdges.contains(successorEdge) == false){
					this.backEdges.add(successorEdge);
					successorEdge.setProperty(BACKEDGE, true);
				}
			}
			
			if(targetBasicBlock.getProperty(VISIT_PROPERTY).equals(UNVISITED))
				searchForBackEdge(targetBasicBlock);
		}
	}
	
	private void conductDepthFirstOrdering(ControlFlowGraph CFG) {
		prepareCFGsForDFSAnalysis(CFG);
		this.globalBasicBlockCount = CFG.getBasicBlocks().size();
		assignOrdering(CFG.getBasicBlock(CFG_ENTRY_BASIC_BLOCK));
		
	}
	
	private void assignOrdering(BasicBlock currentBasicBlock) {
		currentBasicBlock.setProperty(VISIT_PROPERTY, VISITED);
		List<Edge> successorsOfCurrentBasicBlock = currentBasicBlock.getOutgoingEdges(); 
		for(Edge successorEdge : successorsOfCurrentBasicBlock) {
			BasicBlock targetBasicBlock = successorEdge.getTarget();
			
			if(targetBasicBlock.getProperty(VISIT_PROPERTY).equals(UNVISITED))
				assignOrdering(targetBasicBlock);

		}
		currentBasicBlock.setProperty(DFST_BASIC_BLOCK_ID_PROPERTY, this.globalBasicBlockCount);
		--this.globalBasicBlockCount;
	}
	
	private void prepareCFGsForDFSAnalysis(ControlFlowGraph CFG) {
		List<BasicBlock> basicBlocks = CFG.getBasicBlocks();
		for(BasicBlock basicBlock : basicBlocks) {
			basicBlock.setProperty(VISIT_PROPERTY, UNVISITED);
		}
	}
	
	private int getDFSTBasicBlockId(BasicBlock basicBlock) {
		return basicBlock.hasProperty(DFST_BASIC_BLOCK_ID_PROPERTY) ? 
			   (Integer)basicBlock.getProperty(DFST_BASIC_BLOCK_ID_PROPERTY) : 
			   -1;
	}
}
