package com.tetaj.modelgenerator.uppaalmodelgeneratorircfg;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.tetaj.ircfg.*;

import com.jopdesign.wcet.uppaal.model.Location;
import com.jopdesign.wcet.uppaal.model.Template;
import com.jopdesign.wcet.uppaal.model.Transition;

public class MethodTemplate {

	private ControlFlowGraph cfg;
	private Template template;
	private Location executionLocation;
	private Location returnLocation;
	private String uppalName;
	private String uniqueIdentifier;
	private ArrayList<String> nativeCalls;

	public MethodTemplate(String fqMethodName, ControlFlowGraph cfg, String uniqueIdentifier) throws Exception {
		this.cfg = cfg;
		this.uniqueIdentifier = uniqueIdentifier;
		this.setNativeCalls(new ArrayList<String>());
		
		this.template = createTemplate(fqMethodName, cfg);
	}
	
	public Template getTemplate(){
		return this.template;
	}
	
	private Template createTemplate(String fqMethodName, ControlFlowGraph cfg) throws Exception{
		
		Template template = createBasicTemplate(fqMethodName);
		
			BasicBlockModel[] basicBlockModels = createUppaalModelForBasicBlocks();
			
			for(BasicBlockModel basicBlockModel : basicBlockModels){
				if(basicBlockModel.isNativeCall())
					this.nativeCalls.add(basicBlockModel.getNativeCallName());
					
				for(Location location : basicBlockModel.getLocations()){
					template.addLocation(location);
				}
				for(Transition transition : basicBlockModel.getTransitions()){
					template.addTransition(transition);
				}
			}
			connectBasicBlockModels(template, cfg, basicBlockModels);
		
		
		return template;
	}

	private void connectBasicBlockModels(Template template, ControlFlowGraph cfg, BasicBlockModel[] basicBlockModels) throws Exception {
		
		Transition executionTransition = new Transition(this.executionLocation, basicBlockModels[0].getIngoingLocation());
		template.addTransition(executionTransition);
		
		for (BasicBlock basicBlock : (List<BasicBlock>)cfg.getBasicBlocks()) {
			Location outgoingLocation = basicBlockModels[basicBlock.getId()].getOutgoingLocation();
			
			for (Edge edge : basicBlock.getOutgoingEdges()) {
				
				Location incommingLocation = basicBlockModels[edge.getTarget().getId()].getIngoingLocation();
				Transition connectTransition = new Transition(outgoingLocation, incommingLocation);
				//connectTransition.getAttrs().appendUpdate("pm:=pm +1");

				if (edge.isLoopIteration()){
					String loopString = "loop_" + edge.getLoopIterationInfo().getId();
					int loopBound = edge.getLoopIterationInfo().getLoopBound();
						
					connectTransition.getAttrs().appendGuard(loopString + "<" + loopBound);
					connectTransition.getAttrs().appendUpdate(loopString+"++");
					
					if(loopBound!=-1){
						template.appendDeclaration("int[0,"+ loopBound + "] "  + loopString + "=0;");
					}else{
						template.appendDeclaration("int "  + loopString + "=0;");
					}
				}
				else if(edge.isLoopExit()){
					for (LoopInfo loopExitInfo : edge.getLoopExitInfo()) {
						String loopString = "loop_" + loopExitInfo.getId();
						
						connectTransition.getAttrs().appendGuard(loopString+ " == " + loopExitInfo.getLoopBound());
						
						//Progress Measure
//						int strIndex = 0;
//						boolean found=false;
//						for(String str : connectTransition.getAttrs().getUpdates()){
//							if(str.contains("pm:=")){
//								found=true;
//								break;
//							}
//							strIndex++;
//						}
//						if(found){
//							connectTransition.getAttrs().getUpdates().remove(0);
//							connectTransition.getAttrs().appendUpdate("pm:=pm + " + edge.getProperty("pm") +  loopString);
//						}
						connectTransition.getAttrs().appendUpdate(loopString + "=0");
						
					}
				}
				else if (edge.isLoopEntry()) {
					String loopString = "loop_" + edge.getLoopEntryInfo().getId();
					
					connectTransition.getAttrs().appendUpdate(loopString + "=0");
				}
				
				template.addTransition(connectTransition);
			}
			
			//If basic block has no next or target it must be connected to return
			if (basicBlock.getOutgoingEdges().size() == 0){
				Transition connectTransition = new Transition(outgoingLocation,this.returnLocation);
				template.addTransition(connectTransition);
			}
		}
	
	}
	
	private BasicBlockModel[] createUppaalModelForBasicBlocks() {
		BasicBlockModel[] models = new BasicBlockModel[this.cfg.getBasicBlocks().size()];
				
		for(int i=0;i<this.cfg.getBasicBlocks().size();i++){		
			BasicBlockModel basicBlockModel = new BasicBlockModel(this.cfg.getBasicBlock(i), this.uniqueIdentifier);
			models[i] = basicBlockModel;
		}
				
		return models;
	}
	
	private Template createBasicTemplate(String fqMethodName) {
		List<String> templateParameterList = Arrays.asList("");	
		this.uppalName = FQNameHelper.getUppaalFriendlyName(fqMethodName)+"_T";
		Template template = new Template(this.uppalName , templateParameterList);
			
		Location initialLocation = new Location("Idle");
		template.setInitialLocation(initialLocation);
		this.executionLocation = new Location("Executing");
		this.executionLocation.setCommited();
		template.addLocation(this.executionLocation);
		
		Transition initiateTransition = new Transition(initialLocation, this.executionLocation);
		initiateTransition.getAttrs().setSync("invoke_" + this.uniqueIdentifier + "_" + FQNameHelper.getUppaalFriendlyName(fqMethodName) + "?");
		//initiateTransition.getAttrs().appendUpdate("pm:=pm +1");

		template.addTransition(initiateTransition);
		
		this.returnLocation = new Location("Return");
		this.returnLocation.setCommited();
		template.addLocation(this.returnLocation);
		
		Transition returnTransition = new Transition(returnLocation, initialLocation);
		returnTransition.getAttrs().setSync("return_"  + this.uniqueIdentifier + "_"+ FQNameHelper.getUppaalFriendlyName(fqMethodName) + "!");
		//returnTransition.getAttrs().appendUpdate("pm:=pm +1");
		template.addTransition(returnTransition);
		
		return template;
	}

	public String getUppaalName() {
		return this.uppalName;
	}

	public void setNativeCalls(ArrayList<String> nativeCalls) {
		this.nativeCalls = nativeCalls;
	}

	public ArrayList<String> getNativeCalls() {
		return nativeCalls;
	}
	
}
