package com.tetaj.prettyprinter;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;

import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.Instruction;
import com.tetaj.modelgenerator.uppaalmodelgeneratorircfg.FQNameHelper;

import att.grappa.Attribute;
import att.grappa.Edge;
import att.grappa.Graph;
import att.grappa.Node;


public class IRCFGPrettyPrinter {

	private static final String DFST_BASIC_BLOCK_ID_PROPERTY = "dfstvalue";
	
	final String outputFolder;
	private HashMap<Integer, Node> blockToNodeMap = new HashMap<Integer, Node>();
	
	
	public IRCFGPrettyPrinter(final String outputFolder) {
		this.outputFolder = outputFolder;
	}
	
	public void createVisualRepresentation(ControlFlowGraph cfg) throws FileNotFoundException, InterruptedException {
		blockToNodeMap.clear();
		Graph buildedGraph = processSingleCfg(cfg);
		
		File file = new File(this.outputFolder + FQNameHelper.getUppaalFriendlyName(cfg.getMethodName())); 
		FileOutputStream fo = new FileOutputStream(file);
		buildedGraph.printGraph(fo);
		
		try {
			fo.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		convertDotFileToPng(file);
	}
	
	public void createVisualRepresentation(HashMap<String, ControlFlowGraph> cfgs) throws FileNotFoundException, InterruptedException {
		for(ControlFlowGraph cfg : cfgs.values()){
			createVisualRepresentation(cfg);
		}
	}
	
	private void convertDotFileToPng(File file) throws InterruptedException {
		
		String dotCmd = "dot " + file.getAbsolutePath() + " -Tpng -o " + file.getAbsolutePath() + ".png";
		
		try {
			Process p = Runtime.getRuntime().exec(dotCmd);
			p.waitFor();
			p.exitValue();
			p.destroy();
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	

	
	@SuppressWarnings("unchecked")
	private Graph processSingleCfg(ControlFlowGraph cfg) throws FileNotFoundException {
		String graphName = cfg.getSourceFilePath();
		Graph cfgGraph = new Graph(graphName);
		Attribute graphAttr = new Attribute(Attribute.SUBGRAPH, Attribute.LABEL_ATTR, graphName);
		cfgGraph.setAttribute(graphAttr);
			
		for(BasicBlock basicblock : (List<BasicBlock>)cfg.getBasicBlocks()) {
			Node sourceNode = insertNewNodeToGraph(basicblock, cfgGraph);
			
			for(com.tetaj.ircfg.Edge edge : basicblock.getOutgoingEdges()) {
				
				Node targetNode = insertNewNodeToGraph(edge.getTarget(), cfgGraph);
				addEdgeToNode(sourceNode, targetNode, cfgGraph);
			}
		}

		return cfgGraph;
	}
	
	private Node insertNewNodeToGraph(BasicBlock basicblock, Graph cfg) {
		int basicBlockIndex = basicblock.getId();
		if(blockToNodeMap.containsKey(basicBlockIndex)) {
			return blockToNodeMap.get(basicBlockIndex);
		}
		else {
			Node targetNode = new Node(cfg, getNodeName(basicblock));
			Attribute attr = new Attribute(Attribute.NODE, Attribute.LABEL_ATTR, getLabelString(basicblock));
			targetNode.setAttribute(attr);
			blockToNodeMap.put(basicBlockIndex, targetNode);
			return targetNode;
		}
	}
	
	private void addEdgeToNode(Node sourceNode, Node targetNode, Graph cfg) {
		Edge basicBlockEdge = new Edge(cfg, sourceNode, targetNode);
		cfg.addEdge(basicBlockEdge);
	}
	
	private String getNodeName(BasicBlock basicBlock) {
		return Integer.toString(basicBlock.getId());
	}
	
	private String getLabelString(BasicBlock basicBlock) {
		StringBuilder sb = new StringBuilder();
		
		sb.append("Basic Block (" + basicBlock.getId() + ")\r");
		if(basicBlock.hasProperty(DFST_BASIC_BLOCK_ID_PROPERTY)) {
			sb.append("Depth First ordering ID (" + basicBlock.getProperty(DFST_BASIC_BLOCK_ID_PROPERTY) + ")\r");
		}
		int currentLine = 0;
		for(Instruction instruction : basicBlock.getInstructions()){
			
			int instructionLineNum = instruction.getLinenumber();
			
			if(currentLine != instructionLineNum){
				sb.append("Line Number: " + instructionLineNum + ")\r");
				currentLine = instructionLineNum;
			}
			
			sb.append("\t" + instruction.getMnemonic() + " " + instruction.getOperands() + "\r");	
		}
		return sb.toString();
	}
	
	public static void main(String argv[]) {
		/*String[] methods = {"doWhileLoop()V",
							"forLoop()I",
							"foreachLoop()I",
							"whileLoop()V",
							"nestedLoop()V"};
		
		CFGGenerator cfgGenerator;
		ControlFlowGraph cfg;
		IRCFGPrettyPrinter pp;
		ArrayList<String> sourcePaths = AppConfig.getSourcePaths();
		sourcePaths.add("./resources/outsideclasspath/");
		sourcePaths.add("./resources/testappsrc/");
		*/
		
		/*
		 * for eclipse
		 */
		/*for(String method : methods) {
			cfgGenerator = new CFGGenerator("org.testapp.ControlStructures#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.testapp.ControlStructures#" + method);
			pp = new IRCFGPrettyPrinter("./prettyprintedCfgs/eclipse/" + method + "-Eclipse.dot");
			pp.createVisualRepresentation(cfg);
		}
		 */
		
		/*
		 * For javac
		 */
		/*for(String method : methods) {
			cfgGenerator = new CFGGenerator("org.outsideclasspath.ControlStructuresJavac#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.outsideclasspath.ControlStructuresJavac#" + method);
			pp = new PrettyPrinter("./prettyprintedCfgs/javac/" + method + "-Javac.dot");
			pp.createVisualRepresentation(cfg);
		}*/
		
		/*
		 * For GCJ
		 */
		/*for(String method : methods) {
			cfgGenerator = new CFGGenerator("org.outsideclasspath.ControlStructuresGcj#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.outsideclasspath.ControlStructuresGcj#" + method);
			pp = new PrettyPrinter("./prettyprintedCfgs/gcj/" + method + "-GCJ.dot");
			pp.createVisualRepresentation(cfg);
		}*/
		
		/*
		 * For Jikes
		 */
		/*for(String method : methods) {
			if(method.equals("foreachLoop()I")) continue;
			cfgGenerator = new CFGGenerator("org.outsideclasspath.ControlStructuresJikes#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.outsideclasspath.ControlStructuresJikes#" + method);
			pp = new PrettyPrinter("./prettyprintedCfgs/jikes/" + method + "-Jikes.dot");
			pp.createVisualRepresentation(cfg);
		}*/
		
	}
	
}
