package com.tetaj.prettyprinter;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;


import att.grappa.Attribute;
import att.grappa.Edge;
import att.grappa.Graph;
import att.grappa.Node;
import com.jopdesign.libgraph.cfg.ControlFlowGraph;
import com.jopdesign.libgraph.cfg.GraphException;
import com.jopdesign.libgraph.cfg.block.BasicBlock;
import com.jopdesign.libgraph.cfg.statements.stack.StackStatement;
import com.jopdesign.libgraph.struct.TypeException;
import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.cfggenerators.jophelper.CFGGenerator;

public class JopCfgPrettyPrinter {

	final String outputLocation;
	private HashMap<Integer, Node> blockToNodeMap = new HashMap<Integer, Node>();
	
	
	public JopCfgPrettyPrinter(final String outputLocation) {
		this.outputLocation = outputLocation;
	}
	
	public void createVisualRepresentation(ControlFlowGraph cfg) throws FileNotFoundException {
		blockToNodeMap.clear();
		Graph buildedGraph = processSingleCfg(cfg);
		
		FileOutputStream fo = new FileOutputStream(new File(this.outputLocation));
		buildedGraph.printGraph(fo);
		
		try {
			fo.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		convertDotFileToPng();
	}
	
	public void createVisualRepresentation(HashMap<String, ControlFlowGraph> cfgs) throws FileNotFoundException {
		for(ControlFlowGraph cfg : cfgs.values()){
			createVisualRepresentation(cfg);
		}
	}
	
	private void convertDotFileToPng() {
		int filenameBeginIndex = this.outputLocation.lastIndexOf('/') + 1;
		String filename = this.outputLocation.substring(filenameBeginIndex);
		String path = this.outputLocation.substring(0, filenameBeginIndex);
		String strippedFileName = filename.replace(".dot", "");
		
		String dotCmd = "dot " + this.outputLocation + " -Tpng -o " + path + strippedFileName + ".png";
		
		try {
			Runtime.getRuntime().exec(dotCmd);
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	

	
	@SuppressWarnings("unchecked")
	private Graph processSingleCfg(ControlFlowGraph cfg) throws FileNotFoundException {
		String graphName = cfg.getSignature().toString();
		Graph cfgGraph = new Graph(graphName);
		Attribute graphAttr = new Attribute(Attribute.SUBGRAPH, Attribute.LABEL_ATTR, graphName);
		cfgGraph.setAttribute(graphAttr);
			
		for(BasicBlock basicblock : (List<BasicBlock>)cfg.getBlocks()) {
			Node sourceNode = insertNewNodeToGraph(basicblock, cfgGraph);

			for(com.jopdesign.libgraph.cfg.block.BasicBlock.Edge targetEdge : 
				(List<com.jopdesign.libgraph.cfg.block.BasicBlock.Edge>)basicblock.getTargetEdges()) {
				
				BasicBlock targetBasicBlock = targetEdge.getTargetBlock();
				Node targetNode = insertNewNodeToGraph(targetBasicBlock, cfgGraph);
				addEdgeToNode(sourceNode, targetNode, cfgGraph);
			}
			if(basicblock.hasNextBlock()) {
				BasicBlock tempblock = basicblock.getNextBlockEdge().getTargetBlock();
				
				Node targetNode = insertNewNodeToGraph(tempblock, cfgGraph);
				addEdgeToNode(sourceNode, targetNode, cfgGraph);
			}
		}

		return cfgGraph;
	}
	
	private Node insertNewNodeToGraph(BasicBlock basicblock, Graph cfg) {
		int basicBlockIndex = basicblock.getBlockIndex();
		if(blockToNodeMap.containsKey(basicBlockIndex)) {
			return blockToNodeMap.get(basicBlockIndex);
		}
		else {
			Node targetNode = new Node(cfg, getNodeName(basicblock));
			Attribute attr = new Attribute(Attribute.NODE, Attribute.LABEL_ATTR, getLabelString(basicblock));
			targetNode.setAttribute(attr);
			blockToNodeMap.put(basicBlockIndex, targetNode);
			return targetNode;
		}
	}
	
	private void addEdgeToNode(Node sourceNode, Node targetNode, Graph cfg) {
		Edge basicBlockEdge = new Edge(cfg, sourceNode, targetNode);
		cfg.addEdge(basicBlockEdge);
	}
	
	private String getNodeName(BasicBlock basicBlock) {
		return Integer.toString(basicBlock.getBlockIndex());
	}
	
	private String getLabelString(BasicBlock basicBlock) {
		StringBuilder sb = new StringBuilder();
		
		sb.append("Basic Block (" + basicBlock.getBlockIndex() + ")\r\r");
		
		int currentLine = 0;
		for(int i = 0; i < basicBlock.getStackCode().size(); i++){
			
			StackStatement stm = basicBlock.getStackCode().getStackStatement(i);
			int stackStatementLineNum = stm.getLineNumber();
			
			if(currentLine != stackStatementLineNum){
				sb.append("Line Number: " + stackStatementLineNum + ")\r");
				currentLine = stackStatementLineNum;
			}
			
			int opCode = stm.getOpcode();
			if(opCode != -1) {
				sb.append("\t" + Constants.JBC_OPCODE_MNEMONICS[opCode] + " (" + opCode+ ")"+ "\r");
			}		
		}
		return sb.toString();
	}
	
	public static void main(String argv[]) throws FileNotFoundException, GraphException, TypeException {
		String[] methods = {"doWhileLoop()V",
							"forLoop()I",
							"foreachLoop()I",
							"whileLoop()V",
							"nestedLoop()V"};
		
		CFGGenerator cfgGenerator;
		ControlFlowGraph cfg;
		JopCfgPrettyPrinter pp;
		ArrayList<String> sourcePaths = AppConfig.getSourcePaths();
		sourcePaths.add("./resources/outsideclasspath/");
		sourcePaths.add("./resources/testappsrc/");
		
		/*
		 * for eclipse
		 */
		for(String method : methods) {
			cfgGenerator = new CFGGenerator("org.testapp.ControlStructures#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.testapp.ControlStructures#" + method);
			pp = new JopCfgPrettyPrinter("./prettyprintedCfgs/eclipse/" + method + "-Eclipse.dot");
			pp.createVisualRepresentation(cfg);
		}

		/*
		 * For javac
		 */
		/*for(String method : methods) {
			cfgGenerator = new CFGGenerator("org.outsideclasspath.ControlStructuresJavac#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.outsideclasspath.ControlStructuresJavac#" + method);
			pp = new PrettyPrinter("./prettyprintedCfgs/javac/" + method + "-Javac.dot");
			pp.createVisualRepresentation(cfg);
		}*/
		
		/*
		 * For GCJ
		 */
		/*for(String method : methods) {
			cfgGenerator = new CFGGenerator("org.outsideclasspath.ControlStructuresGcj#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.outsideclasspath.ControlStructuresGcj#" + method);
			pp = new PrettyPrinter("./prettyprintedCfgs/gcj/" + method + "-GCJ.dot");
			pp.createVisualRepresentation(cfg);
		}*/
		
		/*
		 * For Jikes
		 */
		/*for(String method : methods) {
			if(method.equals("foreachLoop()I")) continue;
			cfgGenerator = new CFGGenerator("org.outsideclasspath.ControlStructuresJikes#" + method, ".");
			cfg = cfgGenerator.getCfgMap().get("org.outsideclasspath.ControlStructuresJikes#" + method);
			pp = new PrettyPrinter("./prettyprintedCfgs/jikes/" + method + "-Jikes.dot");
			pp.createVisualRepresentation(cfg);
		}*/
		
	}
	
}
