package com.tetaj.modelgenerator.cfganalysis;

import java.io.IOException;
import java.util.HashMap;

import jintgen.gen.PrettyPrinter;

import org.junit.Before;
import org.junit.Test;

import com.tetaj.cfggenerators.AVRToIRCFG;
import com.tetaj.cfggenerators.IIRCFGGenerators;
import com.tetaj.cfggenerators.JBCToIRCFG;
import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.ControlFlowGraphException;
import com.tetaj.ircfg.Edge;
import com.tetaj.prettyprinter.AppConfig;
import com.tetaj.prettyprinter.IRCFGPrettyPrinter;

import static org.junit.Assert.*;

import com.jopdesign.libgraph.cfg.GraphException;
import com.jopdesign.libgraph.struct.TypeException;

public class LoopBoundAnalysisTest {

	@Before
	public void setUp() {
		AppConfig.appendSourcePaths("./resources/testappsrc/:./resources/outsideclasspath/:./resources/AVRToIRCFG/:./resources/AVRToIRCFG/missingloopbounds/:/");
	}	
	
	/*
	 * Following is for the eclipse java compiler
	 * 
	 */
	
	@Test (expected=CfgAnalysisException.class)
	public void testExpectMissingAnnotationExceptionForLoop() throws GraphException, TypeException, IOException, CfgAnalysisException, ControlFlowGraphException {
		
		String methodName = "org.testapp.LoopAnalysisMissing#forLoop()I";
		
		JBCToIRCFG cfgGenerator = new JBCToIRCFG(methodName, ".");
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		loopboundAnalysis.conductAnalysis();
	}
	
	@Test (expected=CfgAnalysisException.class)
	public void testExpectMissingAnnotationExceptionForEach() throws GraphException, TypeException, IOException, CfgAnalysisException, ControlFlowGraphException {
		
		String methodName = "org.testapp.LoopAnalysisMissing#foreachLoop()I";
		
		JBCToIRCFG cfgGenerator = new JBCToIRCFG(methodName, ".");
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		loopboundAnalysis.conductAnalysis();
	}
	
	@Test (expected=CfgAnalysisException.class)
	public void testExpectMissingAnnotationExceptionWhile() throws GraphException, TypeException, IOException, CfgAnalysisException, ControlFlowGraphException {
		
		String methodName = "org.testapp.LoopAnalysisMissing#whileLoop()V";
		
		JBCToIRCFG cfgGenerator = new JBCToIRCFG(methodName, ".");
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		loopboundAnalysis.conductAnalysis();
	}
	
	@Test (expected=CfgAnalysisException.class)
	public void testExpectMissingAnnotationExceptionDoWhile() throws GraphException, TypeException, IOException, CfgAnalysisException, ControlFlowGraphException {
		
		String methodName = "org.testapp.LoopAnalysisMissing#doWhileLoop()V";
		
		JBCToIRCFG cfgGenerator = new JBCToIRCFG(methodName, ".");
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		loopboundAnalysis.conductAnalysis();
	}
	
	@Test
	public void testLoopBoundsNoRecognitionOfIf() throws GraphException, TypeException, IOException, CfgAnalysisException, ControlFlowGraphException {
		
		String methodName = "org.testapp.ControlStructures#ifElseStatement()V";
		
		JBCToIRCFG cfgGenerator = new JBCToIRCFG(methodName, ".");

		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		
		loopboundAnalysis.conductAnalysis();

		for(ControlFlowGraph cfg : cfgtable.values()) {
			for(BasicBlock basicblock : cfg.getBasicBlocks()) {
				for (Edge edge : basicblock.getOutgoingEdges()) {
					assertEquals(false, edge.isLoopIteration());
					assertEquals(false, edge.isLoopExit());
				}
			}
		}
	}
	
	private void testLoopBoundX(String methodName, int entryBlockIndex, int entryEdgeIndex, int exitBlockIndex, int exitEdgeIndex, int loopBound, IIRCFGGenerators cfgGenerator) throws Exception {
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		
		loopboundAnalysis.conductAnalysis();

		ControlFlowGraph cfgForEach = cfgtable.get(methodName);
		
		BasicBlock entryBasicBlock = cfgForEach.getBasicBlock(entryBlockIndex);
		Edge entryEdge = entryBasicBlock.getOutgoingEdges().get(entryEdgeIndex);
		
		assertTrue(entryEdge.isLoopIteration());
		assertFalse(entryEdge.isLoopExit());
		
		BasicBlock exitBasicBlock = cfgForEach.getBasicBlock(exitBlockIndex);
		Edge exitEdge = exitBasicBlock.getOutgoingEdges().get(exitEdgeIndex);
		
		assertFalse(exitEdge.isLoopIteration());
		assertTrue(exitEdge.isLoopExit());		
		
		assertEquals(loopBound, entryEdge.getLoopIterationInfo().getLoopBound());
	}
	
	private void testLoopBoundJBC(String methodName, int entryBlockIndex, int entryEdgeIndex, int exitBlockIndex, int exitEdgeIndex, int loopBound) throws Exception {
		
		JBCToIRCFG cfgGenerator = new JBCToIRCFG(methodName, ".:./resources/outsideclasspath");
		
		/*IRCFGPrettyPrinter pp = new IRCFGPrettyPrinter("./prettyprintedCfgs/ppFilesForTests/" + methodName + ".dot");
		
		pp.createVisualRepresentation(cfgGenerator.generateIRCFG());
		*/
		testLoopBoundX(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, loopBound, cfgGenerator);
	}
	
	private void testLoopBoundAVR(String methodName, String programLocation, int entryBlockIndex, int entryEdgeIndex, int exitBlockIndex, int exitEdgeIndex, int loopBound) throws Exception {
		
		AVRToIRCFG cfgGenerator = new AVRToIRCFG(methodName, programLocation);
		
		/*String tempstr = programLocation.replace("/", "");
	
		IRCFGPrettyPrinter pp = new IRCFGPrettyPrinter("./prettyprintedCfgs/ppFilesForTests/" + tempstr + ".dot");
		
		pp.createVisualRepresentation(cfgGenerator.generateIRCFG());
		*/
		testLoopBoundX(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, loopBound, cfgGenerator);
	}
	
	@Test
	public void testLoopBoundForEach() throws Exception {
		
		String methodName = "org.testapp.ControlStructures#foreachLoop()I";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 2;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 2;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1337);
	}
	
	@Test
	public void testLoopBoundForLoop() throws Exception {
		
		String methodName = "org.testapp.ControlStructures#forLoop()I";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 2;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 2;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1337);
	}
	
	@Test
	public void testLoopBoundWhileLoop() throws Exception {
		
		String methodName = "org.testapp.ControlStructures#whileLoop()V";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 2;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 2;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1337);
	}
	
	@Test
	public void testLoopBoundDoWhileLoop() throws Exception {
		
		String methodName = "org.testapp.ControlStructures#doWhileLoop()V";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 1;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 1;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);

	}
	
	@Test
	public void testLoopBoundWhileContinue() throws Exception {
		
		String methodName = "org.testapp.ControlStructures#whileContinueLoop()V";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 4;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 4;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);

	}	
	
	@Test
	public void testLoopBoundNestedLoop() throws Exception {
		
		String methodName = "org.testapp.ControlStructures#nestedLoop()V";
		
		//Value retrieved using the prettyprinter
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 2;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 2;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);		
		
		/*
		 * inner loop
		 */
		//Value retrieved using the prettyprinter
		entryBlockIndex = 1;
		entryEdgeIndex = 1;
		
		exitBlockIndex = 1;
		exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);
	}
	
	
	/*
	 * Following is for the javac compiler
	 * 
	 */
	
	@Test
	public void testLoopBoundForEachJavac() throws Exception {
		
		String methodName = "org.outsideclasspath.ControlStructuresJavac#foreachLoop()I";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 1;
		int entryEdgeIndex = 0;
		
		int exitBlockIndex = 1;
		int exitEdgeIndex = 1;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1337);

	}
	
	@Test
	public void testLoopBoundForLoopJavac() throws Exception {
		
		String methodName = "org.outsideclasspath.ControlStructuresJavac#forLoop()I";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 1;
		int entryEdgeIndex = 0;
		
		int exitBlockIndex = 1;
		int exitEdgeIndex = 1;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1337);

	}
	
	@Test
	public void testLoopBoundWhileLoopJavac() throws Exception {
		
		String methodName = "org.outsideclasspath.ControlStructuresJavac#whileLoop()V";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 1;
		int entryEdgeIndex = 0;
		
		int exitBlockIndex = 1;
		int exitEdgeIndex = 1;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1337);

	}
	
	@Test
	public void testLoopBoundDoWhileLoopJavac() throws Exception {
		
		String methodName = "org.outsideclasspath.ControlStructuresJavac#doWhileLoop()V";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 1;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 1;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundJBC(methodName, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1337);

	}
	
	@Test
	public void testLoopBoundWhileLoopAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/mainwithwhileloop.elf";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 3;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 3;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);

	}
	
	@Test
	public void testLoopBoundForLoopAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/mainwithforloop.elf";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 4;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 4;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 5);

	}
	
	@Test
	public void testLoopBoundDoWhileLoopAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/mainwithdowhileloop.elf";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 6;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 6;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 2);
	}
	
	@Test
	public void testLoopBoundWhileLoopWithBreakAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/mainwithloopbreak.elf";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 3;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 3;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);
	
		exitBlockIndex = 6;
		exitEdgeIndex = 1;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);
	}
	
	@Test
	public void testLoopBoundWhileLoopWithContinueAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/mainwithloopcontinue.elf";
		
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 3;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 3;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);
	}
	
	@Test
	public void testLoopBoundNestedWhileLoopAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/mainwithnestedwhileloops.elf";
		
		//Value retrieved using the prettyprinter
		//Value retrieved using the prettyprinter
		int entryBlockIndex = 3;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 3;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);		
		
		/*
		 * inner loop
		 */
		//Value retrieved using the prettyprinter
		entryBlockIndex = 6;
		entryEdgeIndex = 1;
		
		exitBlockIndex = 6;
		exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 4);

	}
	
	@Test
	public void testLoopBoundMixedLoopAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/mainwithnestedmixedloops.elf";
		
		//Value retrieved using the prettyprinter
		/*
		 * while loop
		 */
		int entryBlockIndex = 4;
		int entryEdgeIndex = 1;
		
		int exitBlockIndex = 4;
		int exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 1);		
		
		//Value retrieved using the prettyprinter
		/*
		 * for loop
		 */
		entryBlockIndex = 8;
		entryEdgeIndex = 1;
		
		exitBlockIndex = 8;
		exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 2);
		
		//Value retrieved using the prettyprinter
		/*
		 * do-while
		 */
		entryBlockIndex = 14;
		entryEdgeIndex = 1;
		
		exitBlockIndex = 14;
		exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 3);
		
		//Value retrieved using the prettyprinter
		/*
		 * for loop - the last one
		 */
		entryBlockIndex = 10;
		entryEdgeIndex = 1;
		
		exitBlockIndex = 10;
		exitEdgeIndex = 0;
		
		this.testLoopBoundAVR(methodName, programLocation, entryBlockIndex, entryEdgeIndex, exitBlockIndex, exitEdgeIndex, 4);
		
	}
	@Test (expected=CfgAnalysisException.class)
	public void testExpectMissingAnnotationExceptionWhileAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/missingloopbounds/mainwithwhileloopmissing.elf";
		
		AVRToIRCFG cfgGenerator = new AVRToIRCFG(methodName, programLocation);
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		loopboundAnalysis.conductAnalysis();
	}
	
	@Test (expected=CfgAnalysisException.class)
	public void testExpectMissingAnnotationExceptionForAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/missingloopbounds/mainwithforloopmissing.elf";
		
		AVRToIRCFG cfgGenerator = new AVRToIRCFG(methodName, programLocation);
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		loopboundAnalysis.conductAnalysis();
	}
	
	@Test (expected=CfgAnalysisException.class)
	public void testExpectMissingAnnotationExceptionDoWhileAVR() throws Exception {
		
		String methodName = "main";
		String programLocation = "resources/AVRToIRCFG/missingloopbounds/mainwithdowhileloopmissing.elf";
		
		AVRToIRCFG cfgGenerator = new AVRToIRCFG(methodName, programLocation);
		HashMap<String, ControlFlowGraph> cfgtable = cfgGenerator.generateIRCFG();

		LoopBoundAnalysis loopboundAnalysis = new LoopBoundAnalysis(cfgtable);
		loopboundAnalysis.conductAnalysis();
	}
	
	
}
